# -*- coding: utf-8 -*-
"""
Created on Tue Jun 23 13:09:32 2020

@author: benjamin
"""

import numpy as np
import scipy
import scipy.signal
from .audiotools import record_to_file
from .subclasses import Formants, FreqSig, TimeFreq, MFCC
from .prosody import Pitch, Intensity
from .utils import *
from .pyxglos import *
import warnings
with warnings.catch_warnings():
    warnings.filterwarnings("ignore",category=FutureWarning)
    import statsmodels.api as sm
from librosa.core import lpc
import parselmouth
from .filetools import *
import scipy.stats as sst
import json
import numpy.matlib as npmt
import copy
from .ola import *
from .praattools import *
from .transformtools import (transform_formant, transform_formant_from_dist,
                             sparse_regression, get_pros_modif_factor_from_dist,
                             transform_pitch)

class SpeechAudio:

    signal = None
    sampling_frequency = 1
    phonetic_labels = None
    phonetic_instants = None
    utter_id = None

    time_vector = None
    spectrum = None
    spectrogram = None
    pitch = None
    instantaneous_pitch = None
    maximal_voiced_frequency = None
    voiced_signal = None
    unvoiced_signal = None
    spectral_centroid = None
    spectral_spread = None
    skewness = None
    kurtosis = None
    spectral_envelope = None
    formants = None
    zeros = None
    hnr = None
    voicing_quotient = None
    segments = None
    intensity = None
    mfcc = None

    # Constructor method
    def __init__(self, *args):
        nargin = len(args)
        for k in range(0, nargin, 2):
            key, value = args[k:k+2]
            setattr(self, key, value)

    def copy(self):

        return copy.deepcopy(self)

    def computespectrogram(self, nwin=256, win=None, nfft=None, novlp=None):

        if self.sampling_frequency == 1:
            warnings.warn("Warning: the sampling frequency is 1")
        if win is None:
            win = 'hanning'
        if nfft is None:
            nfft = 2**nextpow2(self.sampling_frequency)
        if novlp is None:
            novlp = int(nwin * 3 / 4)

        x = self.signal
        sr = self.sampling_frequency
        nhop = nwin - novlp
        xbuff = buffer(x, nwin, nhop)
        w = window(nwin, win)
        L = int(nfft / 2 + 1)
        fxx = np.arange(L) * sr / nfft

        nFrame = xbuff.shape[1]
        txx = np.arange(0, len(x), nhop) / sr
        txx = txx[:nFrame] + (nwin / 2) / sr

        Sxx = np.zeros((L, nFrame))
        w = np.matlib.repmat(w.reshape(-1,1), 1, nFrame)
        xwin = xbuff * w
        # xwin[xwin==0] = 1e-22
        # xEnv = np.zeros((nfft, nFrame))

        Sxx = np.fft.fft(xwin, n=nfft, axis=0)[:L,:]
        # fxx, txx, Sxx = scipy.signal.spectrogram(self.signal, self.sampling_frequency,
        #                                 win, winLength, novlp, nfft)
        Sxx[Sxx == 0] = np.finfo(type(Sxx[0, 0])).eps
        outputSpectro = TimeFreq('parent', self,
                                 'time_frequency_matrix', Sxx,
                                 'time_vector', txx,
                                 'frequency_vector', fxx,
                                 'window', win,
                                 'overlap', novlp,
                                 'nfft', nfft)

        self.spectrogram = outputSpectro

        return Sxx, fxx, txx

    def computespectrum(self, win='rect', nfft=2048, whole=False):
        N = len(self.signal)
        if win == 'rect':
            w = np.ones(N)
        if win == 'hann':
            w = np.hanning(N)
        if win == 'hamming':
            w = np.hamming(N)

        if self.sampling_frequency == 1:
            warnings.warn("Warning: the sampling frequency is 1")

        sig2fft = self.signal*w
        sp = np.fft.fft(sig2fft, nfft)
        freq = np.fft.fftfreq(nfft, 1/self.sampling_frequency)
        if not whole:
            L = int(nfft/2)
            sp = sp[:L]
            freq = freq[:L]

        outputSpectrum = FreqSig('parent', self,
                                 'values', sp,
                                 'frequency_vector', freq,
                                 'window', w,
                                 'nfft', nfft)

        self.spectrum = outputSpectrum
        return sp, freq

    def computespectralmoments(self, winLength=256,
                               win=None, nfft=None, novlp=None):

        if win is None:
            win = scipy.signal.hanning(winLength)
        if nfft is None:
            nfft = 2**nextpow2(self.sampling_frequency)
        if novlp is None:
            novlp = int(winLength * 3 / 4)

        if self.spectrogram is None:
            self.computespectrogram(winLength, win, nfft, novlp)

        f = self.spectrogram_frequency_vector
        Sxx = abs(self.spectrogram)**2

        nFreq, nFrame = Sxx.shape
        cgs = np.zeros(nFrame)
        spread = np.zeros(nFrame)

        for k in range(nFrame):
            Sxxtmp = Sxx[:, k]
            cgs[k] = np.sum(Sxxtmp * f) / np.sum(Sxxtmp)
            spread[k] = np.sqrt(np.sum((f - cgs[k])**2 * Sxxtmp) / np.sum(Sxxtmp))

        self.spectral_centroid = cgs
        self.spectral_spread = spread

        return cgs, spread

    def getpitch(self, f0min=70, f0max=300, isInterp=True):

        sound = sp2sound(self)
        f0min = max(f0min, int(np.ceil(self.sampling_frequency / len(self.signal) * 3 )))
        f0, tf0 = praat_get_pitch(sound, f0min, f0max)
        pitchOut = Pitch('values', f0, 'time_points', tf0, 'parent', self)

        if isInterp:
            f0 = pitchOut.interpolate()

        self.pitch = pitchOut
        return pitchOut

    def getjitter(self, nwin=5):
        sound = sp2sound(self)
        if self.pitch is None:
            self.getpitch()
        f0min = 0.8 * np.min(self.pitch.values)
        f0max = 1.2 * np.max(self.pitch.values)
        time, jitter =  praat_get_jitter(sound, nwin=nwin, f0min=f0min, f0max=f0max)
        self.pitch.glottal_closure_instants = time
        self.pitch.jitter = jitter

    def changecontour(self, pitchTarget):

        if self.pitch is None:
            inPitch = self.getpitch()
        inPitch = self.pitch
        sound = sp2sound(self)
        sOut = praat_modif_pitch(sound, inPitch.time_points, pitchTarget.values)
        sound2sp(sOut, self)

    def changejitter(self, factor, f0min=70, f0max=300):

        sound = sp2sound(self)
        oldT, pointProcess = praat_get_gci(sound, f0min, f0max)
        pointProcess = praat_remove_points(pointProcess)

        r = 4 * np.random.rand(len(oldT)) - 2
        To = np.abs(np.diff(oldT))
        To = np.insert(To, 0, 0)
        newT = np.sort(oldT + r * factor / 100 * To)

        sOut = praat_change_gci(sound, newT, pointProcess)
        sound2sp(sOut, self)

    def changeshimmer(self, factor, f0min=70, f0max=300):

        sound = sp2sound(self)
        sOut = praat_change_shimmer(sound, factor, f0min=70, f0max=300)
        sound2sp(sOut, self)

    def getintensity(self, len_win=0.03):

        y = self.signal.squeeze()
        sr = self.sampling_frequency
        tsig = np.arange(0, len(y)) / sr

        nwin = int(len_win * sr)
        if np.mod(nwin, 2) > 0:
            nwin = nwin + 1

        energy = []
        s = 0
        for k, smpl in enumerate(y):
            s += smpl**2
            if (k + 1) % nwin == 0:
                energy.append(s)
                s = 0
            elif k == len(y) - 1:
                energy.append(s)

        yOut_lin = np.array(energy) / np.sqrt(nwin)
        yOut_dB = 10 * np.log10(yOut_lin / 2e-5)
        tVec = tsig[::nwin] + len_win / 2
        objOut = Intensity('values', yOut_dB,
                           'time_points', tVec,
                           'parent', self)

        self.intensity = objOut
        return objOut

    def changeintensity(self, intTarget):

        if self.intensity is None:
            intIn = self.getintensity()
        else:
            intIn = self.intensity
        intIn = intIn.interpolate(self)
        intTarget = intTarget.interpolate(self)

        y = self.signal
        ydetrend = y / intIn.db2lin()
        ychange = ydetrend * intTarget.db2lin()

        self.signal = ychange

    def changetilt(self, factor, durWin=0.04, f0min=70, f0max=300):

        y = self.signal
        sr = self.sampling_frequency

        tVec = np.arange(len(y)) / sr

        nwin = int(durWin * sr)
        if np.mod(nwin, 2) > 0:
            nwin = nwin + 1

        novlp, nfft, nhop = ola_param(nwin, novlp=None, sr=sr)
        L = int(nfft / 2 + 1)
        freq_out = np.arange(L) * sr / nfft

        if self.pitch is None:
            self.getpitch(f0min=f0min, f0max=f0max)

        tf0 = self.pitch.time_points
        f0 = self.pitch.values

        f1 = scipy.interpolate.interp1d(tf0, f0, 'linear', fill_value='extrapolate')
        f0 = f1(tVec)

        w = scipy.signal.hann(nwin)
        f0 = np.vstack((np.zeros((nhop, 1)),
                        f0.reshape(-1, 1),
                        np.zeros((nhop, 1)))).squeeze()
        y = np.vstack((np.zeros((nhop, 1)),
                       y.reshape(-1, 1),
                       np.zeros((nhop, 1)))).squeeze()
        yOut = np.zeros_like(y)

        k = 0
        isC = True
        while isC:
            xTmp, idx, isC = check_index(y, k, nwin, nhop)

            f0Tmp = np.median(f0[idx])
            y_fft = np.fft.fft(xTmp, nfft)[:L]

            idxStart = np.abs(freq_out- f0Tmp).argmin()
            bk = y_fft[idxStart:]
            fk = freq_out[idxStart:]

            p_obs = np.polyfit(fk, np.log(np.abs(bk)), 1)
            p1 = np.log(np.abs(abs(bk[0]))) - p_obs[0] * fk[0]
            line_obs = p_obs[0] * freq_out + p1
            p2 = np.log(np.abs(abs(bk[0]))) - factor * p_obs[0] * fk[0]
            line_targ = factor * p_obs[0] * freq_out + p2

            amp_ratio = np.exp(line_targ - line_obs)
            amp_ratio[freq_out <= fk[0]] = 1

            y_modif_fft = y_fft * amp_ratio
            xNew = symmetricifft(y_modif_fft)[:len(idx)]

            yOut[idx] = yOut[idx] + xNew
            k += 1

        yOut = yOut / 1.5

        SpOut = self.copy()
        SpOut.signal = yOut
        SpOut.sampling_frequency = sr

        return SpOut

    def changeduration(self, pts, factor, isSeg=True):

        sound = sp2sound(self)
        if self.phonetic_instants is None:
            isSeg = False

        if isSeg:
            phon_inst_old = copy.deepcopy(self.phonetic_instants)
            phon_inst_new = copy.deepcopy(self.phonetic_instants)
            currSeg = np.zeros(len(phon_inst_old)).astype(int)
            newDur = np.zeros(len(phon_inst_old))
            oldDur = np.zeros(len(phon_inst_old))

            for k in range(len(pts)):
                currPhonInst = phon_inst_old[0][k, :]
                oldDur[0] = phon_inst_old[0][k, 1] - phon_inst_old[0][k, 0]
                newDur[0] = oldDur[0] * factor[k]
                phon_inst_new[0][k, 1] = phon_inst_new[0][k, 0] + newDur[0]
                phon_inst_new[0][k + 1:, :] += newDur[0] - oldDur[0]
                for kinst in range(1,len(phon_inst_old)):
                    newDur[kinst] += newDur[0]
                    oldDur[kinst] += oldDur[0]
                    if currPhonInst[1] >= phon_inst_old[kinst][currSeg[kinst], 1]:
                        phon_inst_new[kinst][currSeg[kinst], 1] = phon_inst_new[kinst][currSeg[kinst], 0] + newDur[kinst]
                        phon_inst_new[kinst][currSeg[kinst]+ 1 :, :] += newDur[kinst] - oldDur[kinst]
                        currSeg[kinst] += 1
                        newDur[kinst] = 0
                        oldDur[kinst] = 0

        sOut = praat_change_duration(sound, pts, factor)
        sound2sp(sOut, self)
        if isSeg:
            self.phonetic_instants = phon_inst_new

    def xglos(self, param=None):

        param = initparam(self, param)

        S = self.signal

        if self.sampling_frequency == 1:
            warnings.warn("Warning: the sampling frequency is 1")

        sr = self.sampling_frequency

        #### Build sliding window for the time signal
        Lw = int(param.win_length)
        w = scipy.signal.hanning(Lw + 1)
        w = w[1:]
        param.hop = param.win_length / param.hop_factor
        a = param.hop

        ## Initialization
        #### zero-padding of the signal at both extremities
        npad = Lw
        Spad = np.concatenate((np.zeros((npad)), S, np.zeros((npad))))
        Nframes = int(np.ceil((len(Spad) - 2 * Lw) / a))

        t = np.arange(1, Nframes + 1) * (a / sr) + Lw / (2 * sr) - npad / sr
        param.t = t

        ## F0
        if self.pitch is None:
            pks = np.zeros((int(param.nfft), Nframes))
            x = np.zeros((Lw, Nframes))
            f0 = np.empty(Nframes)
            f0[:] = np.nan
            param.lfmed = np.ceil(param.fomin*2.5)

            for u in range(Nframes):
                idx1 = int((u+1)*a-1)
                idx2 = int((u+1)*a+Lw-1)
                x[:, u] = Spad[idx1:idx2]
                xtmp = Spad[idx1:idx2]
                pks[:, u] = findpartials(xtmp, sr, np.ones(Lw), param)
                f0[u] = findf0(pks[:, u], sr, param)# Get pitch and detected harmonics
            #            # adapts the length of the median filter
                if f0[u]:
                    param.lfmed = np.ceil(f0[u]*2.5)
        else:
            f0_raw = self.pitch.values
            tfo = self.pitch.time_points
            f1 = scipy.interpolate.PchipInterpolator(tfo, f0_raw)
            f0 = f1(t)

        fk = f0reeval(f0, Spad, sr, param)
        f0 = fk[:, 0]
        mvf = np.max(fk, axis=1)

        # interpolation of f0 and MVF
        tfo = np.arange(len(Spad)) / sr - npad / sr
        f1 = scipy.interpolate.PchipInterpolator(t, f0)
        foint = f1(tfo)
        foint[np.isnan(foint)] = 0
        aiint = np.ones_like(foint)
        f2 = scipy.interpolate.PchipInterpolator(t, mvf)
        mvfint = f2(tfo)
        mvfint[np.isnan(mvfint)] = 0

        # jitter
        if param.jitt or param.shim:
            sfilt = np.zeros((len(Spad), Nframes))
            fkcmat = np.zeros(Nframes)
            fojmat = np.empty((len(Spad), Nframes))
            fojmat[:] = np.nan
            aimat = np.empty((len(Spad), Nframes))
            aimat[:] = np.nan
            kfk = 0
            for u in range(Nframes):
                idx = np.arange((u+1)*a-1, (u+1)*a+Lw)
                if fk[u, 0] > 0:
                    fkc = np.round(fk[u, 0])
                    if any(fkcmat == fkc) == False:
                        bb = param.fcoef * np.cos(2 * np.pi * fkc /
                                                  sr * np.arange(len(param.fcoef)))
                        xwf = scipy.signal.filtfilt(bb, 1, Spad)
                        sfilt[:, kfk] = xwf
                        fkcmat[kfk] = fkc
                        xwf = xwf[idx.astype(int)]
                        kfk += 1
                    else:
                        xwf = sfilt[idx.astype(int), fkcmat == fkc]

                    #fojitt = instfreq(xwf,sr, fk[u,0])
                    fojitt, aijitt, dum = instprop(xwf, sr)
                    aijitt[np.logical_or(fojitt < param.fomin, fojitt > param.fomax)] = 0
                    fojitt[np.logical_or(fojitt < param.fomin, fojitt > param.fomax)] = 0
                    fojmat[idx.astype(int), u] = fojitt
                    aimat[idx.astype(int), u] = fojitt
            foj = np.nanmedian(fojmat, axis=1)
            aij = np.nanmedian(fojmat, axis=1)
            aij[aij == np.nan] = 0
            aij[np.logical_or(foj < param.fomin, foj > param.fomax)] = 0
            aij = sm.nonparametric.lowess(aij,
                                          np.arange(len(aij)),
                                          frac=101/len(aij),
                                          return_sorted=False)
            foj[foj == np.nan] = 0
            foj[np.logical_or(foj < param.fomin, foj > param.fomax)] = 0
            foj = sm.nonparametric.lowess(foj,
                                          np.arange(len(foj)),
                                          frac=101/len(foj),
                                          return_sorted=False)
            foint = foj
            aiint = aij

        Sh = np.zeros(len(Spad))
        Sh2 = np.zeros(len(Spad))

        harm = fk
        harm[harm == 0] = np.nan

        for u in range(Nframes):
            idx = np.arange((u + 1) * a - 1, (u + 1) * a + Lw)
            wh = scipy.signal.hanning(len(idx) + 1)
            wh = wh[1:]
            s = Spad[idx.astype(int)] * wh
            if fk[u, 0] > 0:
                nk = fk[u, :] / fk[u, 0]
                nk = nk[nk > 0]

                if param.shim:
                    aijitt = aij[idx.astype(int)]
                else:
                    aijitt = 1
                hann_window = npmt.repmat((wh * aijitt).reshape((wh.size, 1)), 1, len(nk))
                if param.jitt:
                    fojitt = foj[idx.astype(int)]
                    alp_phase = ((np.cumsum(fojitt).reshape(-1, 1)).dot(nk.reshape(1, -1)))
                    alp = np.exp(1j * 2 * np.pi / sr * alp_phase)
                else:
                    fk_tmp = (nk * fk[u, 0])
                    sig_tmp = np.arange(len(s)).reshape(-1, 1)
                    alp_phase = 1j * 2 * np.pi / sr
                    alp = np.exp(alp_phase * (sig_tmp).dot(fk_tmp.reshape(1, -1)))

                E = alp[:len(s), :] * hann_window
                bk, dum1, dum2, dum3 = np.linalg.lstsq(E, s.reshape(-1, 1))
                Sp = 2 * np.real(E.dot(bk.reshape(-1, 1))).reshape(-1)
            else:
                Sp = np.zeros(len(s))

            Sh[idx.astype(int)] = Sh[idx.astype(int)] + Sp * wh
            Sh2[idx.astype(int)] = Sh2[idx.astype(int)] + s * wh

        frame_bound = 1.5 * param.hop_factor / 4

        deb = npad
        fin = deb + len(S)
        Sh = Sh[deb:fin] / frame_bound
        Sh2 = Sh2[deb:fin] / frame_bound

        foint = foint[deb:fin]
        mvfint = mvfint[deb:fin]

        self.voiced = SpeechAudio('signal', Sh,
                                  'sampling_frequency', sr)
        self.unvoiced = SpeechAudio('signal', Sh2 - Sh,
                                    'sampling_frequency', sr)

        grossPitch = Pitch('values', harm[:, 0], 'time_points', t)
        instPitch = Pitch('values', foint,
                          'time_points', np.arange(len(self.signal)) / self.sampling_frequency)

        self.instantaneous_pitch = instPitch
        self.maximal_voiced_frequency = mvfint
        self.harmonics = harm
        self.instantaneous_periodic_amplitude = aiint[deb:fin]
        self.pitch = grossPitch

        return Sh, Sh2 - Sh, foint, mvfint, harm

    def backgroundnoise(self, Pxx, targetSNR):

        sig = self.signal
        nSig = len(sig)
        sr = self.sampling_frequency

        nwin = 4096
        novlp, nfft, nhop = ola_param(nwin, novlp=None, sr=sr)
        wNoise = np.random.normal(loc=0, scale=1, size=nSig)
        wNoise, backNoise = init_ola(wNoise, nwin, nhop)

        k = 0
        isC = True
        while isC:
            xwin, fft_xwin, isC, idx = spec_frame(wNoise, k, nwin, nhop, nfft)
            backNoise = change_frame(fft_xwin, Pxx, 1, nfft, idx, backNoise,
                                     xwin, isChange=True)

            k += 1

        backNoise = backNoise[nhop:-nhop] / 1.5
        X = np.sqrt(np.var(sig) / np.var(backNoise) / targetSNR)
        backNoise = backNoise * X
        xOut = sig + backNoise
        self.signal = xOut

        return xOut, backNoise

    def computespectralenvelope(self, nwin=512, novlp=None,
                                nfft=None, order=None):

        sr = self.sampling_frequency
        x = self.signal
        if sr == 1:
            warnings.warn("Warning: the sampling frequency is 1")
        if novlp is None:
            novlp = int(nwin * 3 / 4)
        if nfft is None:
            nfft = nwin
        if order is None:
            order = np.ceil(sr / 1000 + 2)

        nhop = nwin - novlp
        xbuff = buffer(x, nwin, nhop)
        w = np.hamming(nwin)
        preemph = [1, 0.63]
        L = int(nfft / 2 + 1)
        fxx = np.arange(L) * sr / nfft

        nFrame = xbuff.shape[1]
        txx = np.arange(0, len(x), nhop) / sr
        txx = txx[:nFrame]

        Sxx = np.zeros((L, nFrame))

        for k in range(nFrame):
            xwin = xbuff[:, k] * w
            xwin = scipy.signal.filter(1, preemph, xwin)
            lpc_xx = lpc(xwin, order)
            w_lpc, h_lpc = scipy.signal.freqz(1, lpc_xx, L)
            Sxx[:, k] = h_lpc

        outputEnvelope = TimeFreq('parent', self,
                                  'time_frequency_matrix', Sxx,
                                  'time_vector', txx,
                                  'frequency_vector', fxx,
                                  'window', w,
                                  'overlap', novlp,
                                  'nfft', nfft)

        self.spectral_envelope = outputEnvelope

        return Sxx, fxx, txx

    def computecepstrumenvelope(self, nwin=1024, novlp=None,
                                order=None, toldB=None):

        if novlp is None:
            novlp = int(nwin * 3 / 4)

        x = self.signal
        sr = self.sampling_frequency
        novlp, nfft, nhop = ola_param(nwin, novlp=novlp)
        xbuff = buffer(x, nwin, nhop)
        w = np.hamming(nwin)
        L = int(nfft / 2 + 1)
        fxx = np.arange(L) * sr / nfft

        nFrame = xbuff.shape[1]
        txx = np.arange(0, len(x), nhop) / sr
        txx = txx[:nFrame] + (nwin / 2) / sr

        Sxx = np.zeros((L, nFrame))
        if self.pitch is None and order is None:
            self.getpitch(70, 400, False)
            self.pitch.timeinterpolate(txx)
        else:
            self.pitch.timeinterpolate(txx)

        w = np.matlib.repmat(w.reshape(-1,1), 1, nFrame)
        xwin = xbuff * w
        xwin[xwin==0] = 1e-22
        Sxx = np.zeros((L, nFrame))

        for k in range(nFrame):
            order = get_order_ceps(self.pitch.values[k], sr)
            Sxx[:,k] = cepstrum(xwin[:,k], sr, order, toldB)

        outputEnvelope = TimeFreq('parent', self,
                                  'time_frequency_matrix', Sxx,
                                  'time_vector', txx,
                                  'frequency_vector', fxx,
                                  'window', w,
                                  'overlap', novlp,
                                  'nfft', nwin)

        self.spectral_envelope = outputEnvelope

        return Sxx, fxx, txx

    def mfcc(self, nwin=512, novlp=None, nfft=1024, nb_filter=24, nb_coeff=12,
              preemph=[1, -0.95], win='hamming', replace=False):
        """ computes the MFCC coefficients """

        if novlp is None:
            novlp = int(nwin * 3 / 4)
        if nfft is None:
            nfft = nwin

        Sp = self.copy()

        if preemph is not None:
            Sp.signal = scipy.signal.lfilter([1], preemph, Sp.signal)
        Sxx, fxx, txx = Sp.computespectrogram(win=win, nwin=nwin, novlp=novlp,
                                              nfft=nfft)
        filters, mels, mel_freqs = mel_filter_bank(sr=self.sampling_frequency,
                                                   fmin=0, fmax=None,
                                                   nb_filter=nb_filter,
                                                   nfft=nfft)

        Pxx = np.abs(Sxx)**2
        energy = np.sum(Pxx, axis=0).reshape(1, -1)
        Pxx_filtered = filters @ Pxx
        Pxx_log = 10 * np.log10(Pxx_filtered)

        # dct_basis = dct(nb_coeff + 1, filter_len=nb_filter)[1:,:]
        dct_basis = dct(nb_coeff + 1, filter_len=nb_filter)
        static_coeff = dct_basis @ Pxx_log
        c0 = np.array([x for x in static_coeff[0,:]])
        if replace:
            static_coeff[0,:] = energy
        dt = delta(static_coeff)
        ddt = delta(dt)
        mfcc_coeff = np.vstack((static_coeff, dt, ddt))

        mfcc = MFCC('coeff', mfcc_coeff, 'time_vector', txx,
                    'mel_vector', mels, 'frequency_vector', mel_freqs,
                    'power_log', Pxx_log, 'nb_coeff', nb_coeff, 'c0', c0,
                    'energy', energy, 'parent', self)
        mfcc.mfcc2melspectro()

        self.mfcc = mfcc

    def addformants(self, tf_add, nwin=512, novlp=None, nfft=None, nargout=1):

        novlp, _, nhop = ola_param(nwin, novlp)
        x2, x_out = init_ola(self.signal, nwin, nhop)

        k = 0
        isC = True
        while isC:
            xwin, fft_xwin, isC, idx = spec_frame(x2, k, nwin, nhop, nfft)
            x_out = change_frame(fft_xwin, tf_add, 1, nfft, idx, x_out, xwin)
            k += 1

        x_out = x_out[nhop:-nhop] / 1.5

        if nargout < 1:
            self.signal = x_out
        else:
            obj_out = self.copy()
            obj_out.signal = x_out

            return obj_out

    def alignread(self, fileName, utt_id=None):

        if fileName.lower().find('.textgrid') > 0:
            labels, start, stop = func_readTextgrid(fileName)
            labels = removequotes(labels)
            instants = []

            for k in range(len(labels)):
                startTmp = np.array(start[k]).reshape(-1, 1)
                stopTmp = np.array(stop[k]).reshape(-1, 1)
                instTmp = np.concatenate((startTmp, stopTmp), 1)
                instants.append(instTmp)

        elif fileName.lower().find('.xml') > 0:
            labels, start, stop = read_xml_segmentation_files(fileName)
            # labels = removequotes(labels)
            instants = []
            for k in range(len(labels)):
                # startTmp = start[k]).reshape(-1, 1)
                # stopTmp = np.array(stop[k]).reshape(-1, 1)
                instTmp = np.array([start[k], stop[k]]).reshape(1,2)
                if instants == []:
                    instants = instTmp
                else:
                    instants = np.vstack((instants, instTmp))
        elif fileName.lower().find('.json') > 0:
            labels, instants = read_json_segmentation_files(fileName)
        elif fileName.lower().find('.whc') > 0:
            labels, instants = read_whc_segmentation_files(fileName)
        elif fileName.lower().find('.seg') > 0:
            if utt_id is None:
                labels, instants = read_seg_segmentation_files(fileName)
            else:
                labels, instants = read_seg_segmentation_files_2(fileName, utt_id)
        elif fileName.lower().find('.ctm') > 0:
            labels, instants = parse_align_file(fileName)           
        else:
            raise FormatError('Unknown file format')

        self.phonetic_labels = labels
        self.phonetic_instants = instants

        return labels, instants

    def sparse_transform(self, A_source, A_target, isFlat=True, isNorm=True, 
                         is_feat_norm=True, is_residual=True, feat_type="lin",
                         n_jobs=1, alpha=1e-3):

        nb_feat, nb_atoms = A_source.shape
        nwin = (nb_feat - 1) * 2
        novlp = int(nwin * 3 / 4)
        novlp, nfft, nhop = ola_param(nwin, novlp)
        x2, x_out = init_ola(self.signal, nwin, nhop)
        Spx2 = SpeechAudio("signal", x2,
                           "sampling_frequency", self.sampling_frequency)
        Sxx, fxx, txx = Spx2.computecepstrumenvelope(nwin, novlp)
        if feat_type == "log":
            features = np.log(Sxx)
        else:
            features = Sxx * 1
        Sxx_target = sparse_regression(features, A_source, A_target, 
                                       is_feat_norm, is_residual, 
                                       n_jobs=n_jobs, alpha=alpha,
                                       feat_type=feat_type)
        if feat_type == "log":
            Sxx_target = np.exp(Sxx_target)
        self.signal = change_cepstrum(x2, Spx2, x_out, nwin, nhop, nfft, txx, isFlat, isNorm,
                    Sxx, Sxx_target)

    def cepstraltransform(self, transOper, nwin=1024,
                          novlp=None, sr=8000, freq_out=None,
                          method='linear', nargout=1, feat_type='ceps'):

        novlp, nfft, nhop = ola_param(nwin, novlp)
        x2, x_out = init_ola(self.signal, nwin, nhop)

        sr_in = self.sampling_frequency
        if sr_in != sr:
            self.speechresample(sr)

        x2, x_out = init_ola(self.signal, nwin, nhop)
        Spx2 = SpeechAudio('signal', x2, 'sampling_frequency', sr)
        if feat_type == 'ceps':
            Sxx_in, fxx, txx = Spx2.computecepstrumenvelope(nwin, novlp)
        elif feat_type == 'mfcc':
            L = int(nfft / 2 + 1)
            freq_oo = np.arange(L) * self.sampling_frequency / nfft
            nb_coeff = transOper.nb_coeff
            nb_filter = len(transOper.mel_vector) - 2
            S_tmp = transOper.mfcc2melspectro(freq_oo)
            Sxx_trans = S_tmp.time_frequency_matrix

            Spx2.mfcc(nwin=nwin, novlp=novlp,
                      nb_filter=nb_filter, nb_coeff=nb_coeff)
            S_tmp_2 = Spx2.mfcc.mfcc2melspectro(freq_oo)
            Sxx_in = S_tmp_2.time_frequency_matrix

            fxx = S_tmp_2.frequency_vector
            txx = S_tmp_2.time_vector

        if method.lower() == 'dnn':
            """ DNN-based modification is in progress """
            pass
            # if freq_out is not None:
            #     if len(freq_out) != len(fxx) and not (freq_out==fxx).all():
            #         self.spectral_envelope.interpolate(freq_out, axis='freq')
            #         isChange = True
            #     Sxx_in = np.log(self.spectral_envelope.time_frequency_matrix)
            #     Sxx_trans = np.exp(predict(transOper, Sxx_in.T,
            #                                   train_domain='frequency')).T
            # else:
            #     Sxx_in, Sxx_trans = predict(transOper, Sxx_in,
            #                                   train_domain='time')
        else:
            isChange = False
            # if freq_out is not None:
            #     if not (freq_out==fxx):
            #         self.spectral_envelope.interpolate(freq_out, axis='freq')
            #         isChange = True
            #     Sxx_in = self.spectral_envelope.time_frequency_matrix
            if method.lower() == 'linear':
                Sxx_in, Sxx_trans = ceps_transform(Sxx_in, transOper, nwin, novlp,
                                               fxx, freq_out, method)

            if method.lower() == 'target' and feat_type != 'mfcc':
                Sxx_trans = transOper
            if isChange:
                a = TimeFreq('time_frequency_matrix', Sxx_trans,
                            'frequency_vector', freq_out)
                a.interpolate(fxx, axis='freq')
                Sxx_trans = a.time_frequency_matrix
                a = TimeFreq('time_frequency_matrix', Sxx_in,
                            'frequency_vector', freq_out)
                a.interpolate(fxx, axis='freq')
                Sxx_in = a.time_frequency_matrix
            if Sxx_trans.shape[1] < Sxx_in.shape[1]:
                a = TimeFreq('time_frequency_matrix', Sxx_trans,
                            'frequency_vector', freq_out,
                            'time_vector', txx[1:-1] - nhop / sr)
                a.interpolate(txx - nhop / sr, axis='time')
                Sxx_trans = a.time_frequency_matrix
        k = 0
        isC = True

        while isC:
            xwin, fft_xwin, isC, idx = spec_frame(x2, k, nwin, nhop, nfft)

            x_out = change_frame(fft_xwin, Sxx_trans[:,k], Sxx_in[:,k], nfft, idx, x_out,
                                     xwin, isChange=True, isNorm=False)
            k, isC = check_frame(k, isC, maxFrame=Sxx_in.shape[1])
        x_out = x_out[nhop:-nhop] / 1.5

        if nargout == 1:
            obj_out = self.copy()
            obj_out.signal = x_out
            if sr_in != sr:
                obj_out.speechresample(sr_in)
            return obj_out
        else:
            self.signal = x_out
            self.sampling_frequency = sr
            if sr_in != sr:
                self.speechresample(sr_in)

    def createdisto(self, dist_threshold=0.8):

        x = np.zeros_like(self.signal)
        x[:] = self.signal
        xMax = max(abs(x))*dist_threshold
        x[x > xMax] = xMax
        x[x < -xMax] = -xMax
        self.signal = x


    def exportsegment2json(self, jsonFileName):

        if self.phonetic_labels is None:
            raise SegError('No segmentation: cannot export empty data')

        nTiers = len(self.phonetic_labels)
        idjson = {}
        idjson['tiers'] = []
        for k in range(nTiers):
            lbl = self.phonetic_labels[k]
            inst = self.phonetic_instants[k]
            nEvents = len(lbl)
            segStart = []
            segStop = []
            events = []
            for kE in range(nEvents):
                segStart = "%.3f" %(inst[kE, 0])
                segStop = "%.3f" %(inst[kE, 1])
                events.append({'segment_name': lbl[kE],
                               'segment_start': segStart,
                               'segment_stop': segStop})
            idjson['tiers'].append({'events': events})

        fileJSON = jsonFileName.replace('.json', '') + '.json'
        with open(fileJSON, 'w') as f:
            json.dump(idjson, f)

    def formantestimate(self, nwin=512, novlp=None, ord_filt=25):

        sr = self.sampling_frequency
        novlp, _, nhop = ola_param(nwin, novlp)
        snd = sp2sound(self)
        time_vec = np.arange(0, len(self.signal) + nhop, nhop) / sr
        form_out, bw_out, txx = praat_get_formants(snd, nhop, nwin)
        for k in range(4):
            fTmp = np.squeeze(form_out[k, :])
            fTmp = fill_nan_and_zero(fTmp)
            bwTmp = np.squeeze(bw_out[k, :])
            bwTmp = fill_nan_and_zero(bwTmp)

            if ord_filt is not None:
                if ord_filt > 0:
                    num_frames = len(fTmp)
                    if num_frames >= ord_filt:
                        fTmp = sm.nonparametric.lowess(fTmp,
                                                       np.arange(len(fTmp)),
                                                       frac=ord_filt/len(fTmp),
                                                       return_sorted=False)

                        bwTmp = sm.nonparametric.lowess(bwTmp,
                                                        np.arange(len(fTmp)),
                                                        frac=ord_filt/len(fTmp),
                                                        return_sorted=False)
                    else:
                        pass
            form_out[k, :] = fTmp
            bw_out[k, :] = bwTmp

        damp = bw_out * np.pi
        formantObject = Formants('frequency', form_out,
                                 'bandwidth', bw_out,
                                 'damping', damp,
                                 'amplitude', np.ones_like(form_out),
                                 'time_vector', txx,
                                 'parent', self,
                                 'isForm', True)

        formantObject.interpolate(time_vec)
        formantObject.getcontour()
        self.formants = formantObject

        return formantObject

    def formantshifting(self, formTarget, nwin=512,
                        novlp=None, nfft=None, tf_add=1):

        novlp, nfft, nhop = ola_param(nwin, novlp)
        x2, x_out = init_ola(self.signal, nwin, nhop)
        sr = self.sampling_frequency

        Sxx, fxx, txx = self.computecepstrumenvelope(nwin, novlp)
        if tf_add is not None:
            tf_add.interpolate(fxx)
            tf_resp = tf_add.values
        else:
            tf_resp = 1

        nF = len(formTarget)
        [x.interpolate(txx) for x in formTarget]
        isNewZero = False
        if nF == 1:
            formTarget = formTarget[0]
            zeroTarget = None
        if nF == 2:
            zeroTarget = formTarget[1]
            formTarget = formTarget[0]
            isNewZero = True

        newDamp = formTarget.damping
        newFreq = formTarget.frequency
        newZ = np.exp(2 * np.pi * 1j * newFreq / sr) * np.exp(-newDamp / sr)
        newZ = np.concatenate((newZ, np.conj(newZ)), 0)
        if isNewZero:
            newZeroFreq = zeroTarget.frequency
            newZeroDamp = zeroTarget.damping
            newZero = np.exp(2 * np.pi * 1j * newZeroFreq / sr) * np.exp(-newZeroDamp / sr)
            newZero = np.concatenate((newZero, np.conj(newZero)), 0)
        else:
            newZero = np.zeros_like(newZ)

        k = 0
        isC = True

        while isC:
            xwin, fft_xwin, isC, idx = spec_frame(x2, k, nwin, nhop, nfft)
            if (newFreq[:,k] != 0).all():
                _, h_new = z2freqz(newZ[:, k], newZero[:, k], nfft=nfft)
            else:
                h_new = 1
            h_orig = Sxx[:,k]
            x_out = change_frame(fft_xwin, h_new, h_orig,
                                         nfft, idx, x_out, xwin=xwin,
                                         isChange=(newFreq[:,k] != 0).all())

            k, isC = check_frame(k, isC, maxFrame=newFreq.shape[1])

        x_out = x_out[nhop:-nhop] / 1.5
        x_out[np.isnan(x_out)] = 0

        self.signal = x_out
        self.formants = formTarget
        
    def change_prosody(self, dil_fact=1, trans_fact=1, dur_fact=1, dist=None):

        if self.pitch is None:
            self.getpitch()
        if dist is not None:
            trans_fact, dil_fact, dur_fact = get_pros_modif_factor_from_dist(dist)
            
        if dil_fact != 1 and trans_fact != 1:
            pitch_output = transform_pitch(self.pitch, trans_fact, dil_fact)
            self.changecontour(pitch_output)

        if dur_fact != 1:
            pts = [0, len(self.signal)/ self.sampling_frequency]
            factor = np.ones(3) * dur_fact
            self.changeduration(pts, factor)

    def frequency_warping(self, dil_fact, trans_fact, dist=None, ord_filt=None,
                      nwin=512, novlp=None, isNorm=True, isFlat=True):

        novlp, nfft, nhop = ola_param(nwin, novlp)
        x2, x_out = init_ola(self.signal, nwin, nhop)
        Spx2 = SpeechAudio('signal', x2,
                           'sampling_frequency', self.sampling_frequency)
        Sxx, fxx, txx = Spx2.computecepstrumenvelope(nwin, novlp)
        form_in = Spx2.formantestimate(nwin, novlp, ord_filt=ord_filt)
        form_in.interpolate(txx)

        if dist is None:
            form_output = transform_formant(form_in, dil_fact, trans_fact)
        else:
            form_output = transform_formant_from_dist(form_in, dist)
        form_output.interpolate(txx)

        nb_frame = Sxx.shape[1]
        xx, yy = np.meshgrid(np.arange(0, nb_frame), fxx)
        h_new = dfw_matrix(np.abs(Sxx), fxx,
                           np.sort(form_output.frequency, axis=0),
                           form_in.frequency)

        self.signal = change_cepstrum(x2, Spx2, x_out, nwin, nhop, nfft, txx,
                                      isFlat, isNorm, Sxx, h_new)
        self.formantestimate(nwin, novlp, ord_filt=None)
        self.formants.interpolate(txx)

    def time_warping(self, inst_target, tiers=0):
        """ dynamic time warping of phonemes to match the specified target """
        if inst_target.shape[0] != self.phonetic_instants[tiers].shape[0]:
            raise ValueError('The number of phonemes does not match')

        len_input = len(self.signal) / self.sampling_frequency
        inst_input = self.phonetic_instants[0]
        if inst_input[-1,1] < len_input:
            inst_input[-1,1] = len_input

        dur_input =  inst_input[:,1]- inst_input[:,0]
        dur_target = inst_target[:, 1] - inst_target[:, 0]

        factor = dur_target / dur_input
        factor[np.isnan(factor)] = 0
        factor[np.isinf(factor)] = 0

        pts = inst_input[:-1,1]
        self.changeduration(pts, factor, isSeg=True)

    def remove_silence(self, side="both"):
        y = self.signal
        sr = self.sampling_frequency
        lbl = self.phonetic_labels[0]
        inst = self.phonetic_instants[0]

        if lbl[0] == "." and side in ["both", "l"]:
            delay = inst[1,0]
            idx_st = int(delay * sr)
            y = y[idx_st:]
            lbl = lbl[1:]
            inst = inst[1:, :] - delay
            if len(self.phonetic_labels) > 1:
                if self.phonetic_labels[1][0] ==  "[silence]":
                    self.phonetic_labels[1] = self.phonetic_labels[1][1:]
                    self.phonetic_instants[1] = self.phonetic_instants[1][1:, :] - delay
        if lbl[-1] == "." and side in ["both", "r"]:
            idx_st = int(inst[-1,0] * sr)
            y = y[:idx_st]
            lbl = lbl[:-1]
            inst = inst[:-1,:]
            if len(self.phonetic_labels) > 1:
                if self.phonetic_labels[1][-1] ==  "[silence]":
                    self.phonetic_labels[1] = self.phonetic_labels[1][:-1]
                    self.phonetic_instants[1] = self.phonetic_instants[1][:-1, :]

        self.phonetic_labels[0] = lbl
        self.phonetic_instants[0] = inst
        self.signal = y

    def savespeech(self, fileName, norm=None):
        """ saves speech into an audio file """

        if norm is None:
            norm = 'notnormalized'
        if norm is type(str):
            norm = norm.lower()

        sig2save = self.signal
        if norm == 'normalized':
            maxSig = max(abs(sig2save)) * 1.1
        if norm == 'notnormalized':
            maxSig = 1
        if type(norm) is not str:
            maxSig = norm

        sig2save = sig2save / maxSig * 32768
        record_to_file(fileName, sig2save.astype('i2'),
                       2, self.sampling_frequency)

    def segmentspeech(self, tiers=0, isSeg=True):

        if self.phonetic_labels is None:
            raise ValueError('Speech audio object is not segmented, please perform segmentation')

        nTiers = len(self.phonetic_labels)
        ph_label = self.phonetic_labels[tiers]
        ph_ist = self.phonetic_instants[tiers]
        x = self.signal
        sr = self.sampling_frequency
        if self.pitch is not None:
            f0 = self.pitch.values
            tf0 = self.pitch.time_points
        else:
            f0 = None
            tf0 = None
        formants = self.formants
        segments = []

        for k in range(len(ph_label)):
            phon_tmp = SpeechSegment('phoneme_label', ph_label[k])
            phon_tmp.getphoninfo()
            phon_tmp.sampling_frequency = sr
            inst_tmp = ph_ist[k, :]
            phonSegLbl = []
            phonSegInst = []
            for kT in range(tiers + 1):
                lblTiers = self.phonetic_labels[kT]
                instTiers = self.phonetic_instants[kT]
                idxStart = [x for x in range(instTiers.shape[0])
                            if instTiers[x,0] >= inst_tmp[0] and instTiers[x,1] <= inst_tmp[1]]
                lblTmp = [lblTiers[x] for x in idxStart]
                momTmp = instTiers[idxStart,:]
                momTmp = momTmp - momTmp[0,0]
                phonSegLbl.append(lblTmp)
                phonSegInst.append(momTmp)

            idx_sig = [x for x in range(int(inst_tmp[0] * sr),
                                        int(inst_tmp[1] * sr)+1) if x < len(self.signal)]
            phon_tmp.signal = x[np.unique(idx_sig)]

            if f0 is not None and isSeg:
                f02 = [f0[x] for x in range(len(f0)) if (tf0[x] >= inst_tmp[0]
                                                         and tf0[x] <= inst_tmp[1])]
                tf02 = [x for x in tf0 if (x >= inst_tmp[0] and x <= inst_tmp[1])]
                phon_tmp.gross_fundamental_frequency = f02
                phon_tmp.gross_f0_time_vector = tf02

            if formants is not None and isSeg:
                tform = formants.time_vector
                idx_form = [x for x in range(len(tform)) if (tform[x] >= inst_tmp[0]
                                                             and tform[x] <= inst_tmp[1])]
                phon_tmp.formants = Formants('frequency',
                                             formants.frequency[:, idx_form],
                                             'bandwidth', formants.bandwidth[:, idx_form],
                                             'damping', formants.bandwidth[:, idx_form],
                                             'amplitude', formants.amplitude[:, idx_form],
                                             'time_vector', formants.time_vector[idx_form])
            phon_tmp.phonetic_labels = phonSegLbl
            phon_tmp.phonetic_instants = phonSegInst
            segments.append(phon_tmp)

        for k in range(len(segments)):
            if k == 0:
                segments[k].following_phoneme = segments[k+1]
            elif k == len(segments)-1:
                segments[k].previous_phoneme = segments[k-1]
            else:
                segments[k].following_phoneme = segments[k+1]
                segments[k].previous_phoneme = segments[k-1]

        return segments

    def simulmicdisto(self, phonemes, coeff, tiers=1, dist_thresh=1.2):

        if self.phonetic_labels is None:
            raise ValueError('No phonetic segmentation. Please perform speech-text alignment')
        labels = self.phonetic_labels[tiers-1]
        instants = self.phonetic_instants[tiers-1]

        x_modif = self.signal.copy()
        max_sig = np.max(np.abs(x_modif)) * dist_thresh
        sr = self.sampling_frequency

        for k in range(len(phonemes)):
            x_modif = speechmodif(x_modif, sr, phonemes[k],
                                  coeff[k], labels, instants)

        self.signal = x_modif

        return self

    def speechresample(self, sro):

        sr = self.sampling_frequency;
        if sr != sro:
            sig = self.signal;
            num = int(sro * len(sig) / sr)
            sig_output = scipy.signal.resample(sig, num)
            self.signal = sig_output;
            self.sampling_frequency = sro;
        else:
            print('Warning : The output sampling frequency '
                  'is the same as the input. Nothing''s done!')

    def computeVQ(self, nwin=1024, novlp=None, nfft=None, meth="rms"):

        if self.voiced is None or self.unvoiced is None:
            raise ValueError("No separation has been done. Please run X-GLOS")

        sr = self.sampling_frequency;
        tvec = np.arange(0, len(self.signal)) / sr

        if novlp is None:
            novlp = int(nwin * 3 / 4)
        if nfft is None:
            nfft = 2**nextpow2(sr)

        V = self.voiced
        U = self.unvoiced

        if meth.lower() == "spectro":
            Sxx_V, fxx, txx = V.computespectrogram(nwin, novlp=novlp, nfft=nfft)
            Sxx_U, _, _ = U.computespectrogram(nwin, novlp=novlp, nfft=nfft)

            Den = np.sum(abs(Sxx_U)**2, 0) + np.sum(abs(Sxx_V)**2, 0)
            grossVQ = np.sum(abs(Sxx_V)**2, 0) / Den
            grossHNR = 20*np.log10(abs((np.sum(abs(Sxx_V)**2, 0))/
                                       (np.sum(abs(Sxx_U)**2, 0))))
        elif meth.lower() == "rms":
                nhop = nwin - novlp
                Vbuff = buffer(V.signal, nwin, nhop)
                Ubuff = buffer(U.signal, nwin, nhop)
                nFrame = Vbuff.shape[1]
                txx = np.arange(0,len(self.signal) - 1, nhop) / sr
                txx = txx[:nFrame] + nwin / sr / 2

                grossVQ = np.zeros(nFrame)
                grossHNR = np.zeros(nFrame)
                w = scipy.signal.hann(nwin)

                for k in range(nFrame):
                    xpt = Vbuff[:,k] * w
                    xpt = xpt - np.mean(xpt)
                    xnt = Ubuff[:,k] * w
                    xnt = xnt - np.mean(xnt)
                    grossVQ[k] = ((np.linalg.norm(xpt,2)**2)/
                                  (np.linalg.norm(xpt,2)**2+np.linalg.norm(xnt,2)**2))**2
                    grossHNR[k] = ((np.linalg.norm(xpt,2)**2)/(np.linalg.norm(xnt,2)**2))**2
        else:
             raise ValueError("The method to compute VQ does not exist. Please choose either spectro or rms")

        VQ = lininterp1(txx, grossVQ, tvec) * 100
        VQ[VQ < 0] = 0
        VQ[VQ > 100] = 100
        HNR = lininterp1(txx, grossHNR, tvec)

        self.voicing_quotient = VQ
        self.hnr = HNR

        if self.time_vector is None:
            self.time_vector = tvec

        return VQ, HNR
class SpeechSegment(SpeechAudio):

    phoneme_label = None
    phonetic_class = None
    voiced = None
    previous_phoneme = None
    following_phoneme = None
    context = None
    mean_formants = None
    mean_fundamental_frequency = None

    mean_hnr = None
    mean_vq = None
    dispersion_formants = None
    dispersion_fundamental_frequency = None
    dispersion_hnr = None
    dispersion_vq = None
    jitter = None
    shimmer = None
    devoing_index = None

    def __init__(self, *args):
        nargin = len(args)
        for k in range(0, nargin, 2):
            key, value = args[k:k+2]
            setattr(self, key, value)

    def getdevoicing(self):

        phm1 = self.previous_phoneme
        if phm1 is None:
            phm1 = SpeechSegment('voicing_quotient', 0)

        if phm1.phoneme_label == '#':
            phm1.voicing_quotient = 0

        php1 = self.following_phoneme
        if php1 is None:
            php1 = SpeechSegment('voicing_quotient', 0)

        if php1.phoneme_label == '#':
            php1.voicing_quotient = 0

        if ((self.voicing_quotient is None) or
                (phm1.voicing_quotient is None) or
                (php1.voicing_quotient is None)):
            raise PhoneError('Phonemes has no voicing_quotient, '
                             'please compute the voicing quotient')

        maxm1 = max(phm1.voicing_quotient)
        maxp1 = max(php1.voicing_quotient)
        minvq = min(self.voicing_quotient)
        maxadj = max(maxm1, maxp1)
        devoicingIndex = maxadj - minvq
        self.devoicing_index = devoicingIndex

        return devoicingIndex

    def getphoninfo(self):

        ph = self.phoneme_label
        oral_vowel_dict = ['a', 'e', 'i', 'o', 'u', 'y', 'A', 'E', 'O', 'I',
                           'U', 'Y', '}', '2', '@', '6', '7', '8', '9', 'V',
                           '&', '3', 'Q', '{', 'a/', 'e/', 'i/', 'o/', 'u/',
                           'y/', 'A/', 'E/', 'O/', 'I/', 'U/', 'Y/', '}/',
                           '2/', '@/', '6/', '7/', '8/', '9/', 'V/', '&/', '3/',
                           'Q/', '{/', '^']
        nasal_vowel_dict = ['a~', 'e~', 'i~', 'o~', 'u~', 'y~', 'A~', 'E~',
                            'O~', 'I~', 'U~', 'Y~', '}~', '2~', '@~', '8~',
                            '7~', 'V~', '&~', '6~', '3~', '3\~', 'Q~', '{~', '@\~']
        voiced_fricative = ['z', 'Z', 'v']
        voiceless_fricative = ['s', 'S', 'f', 'C', 'X', 'x']
        voiced_plosive = ['b', 'd', 'g', '_b', 'b_', '_d', 'd_', 'g_', '_g']
        voiceless_plosive = ['p', '_p', 'p_', 't', 't_', '_t', 'k', 'k_', '_k']
        liquids = ['l', 'r', 'L', 'R']
        nasal_consonants = ['m', 'n']

        if ph in oral_vowel_dict:
            self.phonetic_class = 'oral vowel'
            self.voiced = True
        if ph in nasal_vowel_dict:
            self.phonetic_class = 'nasal vowel'
            self.voiced = True
        if ph in voiced_fricative:
            self.phonetic_class = 'fricative'
            self.voiced = True
        if ph in voiceless_fricative:
            self.phonetic_class = 'fricative'
            self.voiced = False
        if ph in voiced_plosive:
            self.phonetic_class = 'plosive'
            self.voiced = True
        if ph in voiceless_plosive:
            self.phonetic_class = 'plosive'
            self.voiced = False
        if ph in liquids:
            self.phonetic_class = 'liquid'
            self.voiced = True
        if ph in nasal_consonants:
            self.phonetic_class = 'nasal_consonant'
            self.voiced = True

    def getstats(self, meth='central'):

        if meth.lower() not in ['central', 'mean', 'median']:
            raise MethError('Method is not recongnized. Please choose either mean or median')

        if meth.lower() == 'mean':
            self.mean_formants = np.mean(self.formants, axis=1)
            self.dispersion_formants = np.std(self.formants, axis=1)
            self.mean_fundamental_frequency = np.mean(self.gross_fundamental_frequency)
            self.dispersion_fundamental_frequency = np.std(self.gross_fundamental_frequency)
            self.mean_hnr = np.mean(self.hnr)
            self.dispersion_hnr = np.std(self.hnr)
            self.mean_vq = np.mean(self.voicing_quotient)
            self.dispersion_vq = np.std(self.voicing_quotient)
        if meth.lower() == 'median':
            self.mean_formants = np.median(self.formants, axis=1)
            self.dispersion_formants = sst.median_abs_deviation(self.formants, axis=1)
            self.mean_fundamental_frequency = np.median(self.gross_fundamental_frequency)
            self.dispersion_fundamental_frequency = \
                sst.median_abs_deviation(self.gross_fundamental_frequency)
            self.mean_hnr = np.median(self.hnr)
            self.dispersion_hnr = sst.median_abs_deviation(self.hnr)
            self.mean_vq = np.median(self.voicing_quotient)
            self.dispersion_vq = np.median_abs_deviation(self.voicing_quotient)
        if meth.lower() == 'central':
            numFor = self.formants.shape[1]
            isOdd = np.mod(numFor, 2)
            if isOdd:
                self.mean_formants = self.formants[:, numFor / 2 - 0.5]
            else:
                self.mean_formants = \
                    np.mean(self.formants[:, numFor / 2 - 1:numFor / 2 + 1], axis=1)
            numFor = len(self.gross_fundamental_frequency)
            isOdd = np.mod(numFor, 2)
            if isOdd:
                self.mean_fundamental_frequency = self.gross_fundamental_frequency[numFor / 2 - 0.5]
            else:
                self.mean_fundamental_frequency = \
                    np.mean(self.gross_fundamental_frequency[numFor / 2 - 1:numFor / 2 + 1])
            if self.hnr != []:
                numFor = len(self.voicing_quotient)
                isOdd = np.mod(numFor, 2)
                if isOdd:
                    self.mean_hnr = self.hnr[numFor / 2 - 0.5]
                    self.mean_vq = self.voicing_quotient[numFor / 2 - 0.5]
                else:
                    self.mean_hnr = np.mean(self.hnr[numFor / 2 - 1:numFor / 2 + 1])
                    self.mean_vq = np.mean(self.voicing_quotient[numFor / 2 - 1:numFor / 2 + 1])

def sound2sp(x, sp=None):

    if sp is not None:
        # spOut = sp.copy()
        sp.signal = copy.deepcopy(x.values.squeeze())
        sp.sampling_frequency = x.sampling_frequency
        spOut = sp.copy()
        # sp = spOut.copy()
    else:
        spOut = SpeechAudio('signal', x.values.squeeze(),
                         'sampling_frequency', x.sampling_frequency)
    return spOut

def sp2sound(x):

    sound = parselmouth.Sound(x.signal)
    sound.sampling_frequency = x.sampling_frequency

    return sound



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Dec 18 19:53:24 2018

@author: benjamin
"""

import numpy as np
from .utils import forcecompute

class Oscillator: #  < Network
    # Waveguide is a class that defines a vocal tract waveguide
    #### Properties
    ##### Input
    rest_position = None
    x_position = None
    up_height = None
    down_height = None
    initial_length = None
    initial_mass = None
    initial_stiffness = None
    initial_coupling_stiffness = None
    initial_damping = None
    Qr = None
    Qrc = None
    contact_damping = None
    contact_stiffness = None
    fundamental_frequency = None
    low_frequency_abduction = None
    partial_abduction = None
    model = 'ishi'
    length = None
    width = None
    etha_k1 = None
    etha_k2 = None
    etha_h1 = None
    etha_h2 = None

    # Instantaneous Output
    output_length = None
    isContact = None
    height_vector = None
    mass = None
    stiffness = None
    damping = None
    coupling_stiffness = None
    bernoulli = None
    resistance = None
    inductance = None
    separation_height = None
    mass_position = None
    mass_position_nm1 = None
    mass_position_nm2 = None
    inst_flow = None
    flow_nm1 = 0
    upstream_pressure = None
    downstream_pressure = None
    applied_force = None
    Qu1 = 0
    Qu2 = 0
#         DC_flow = None;

    # Output matrix
    height_vector_output = None
    mass_position_matrix = None
    mass_matrix = None
    stiffness_matrix = None
    damping_matrix = None
    coupling_stiffness_matrix = None

     # Constructor method
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    # Import methods
    def updateimpedance( self, Const, aChink, ain, ac, lc):
        # Update the glottis impedance

        rho = Const.rho
        co_mu = Const.mu
        sep = Const.sep
        armin = Const.amin
        kiter = Const.k_iter
        T = Const.T

        xvec = self.x_position
        lg = self.length
        selfmod = self.model
        yt = self.mass_position

        agp = self.oscillation_factor[kiter]
        h1i = np.sum(yt[:,0]*agp)+self.low_frequency_abduction[kiter]/lg
        h2i = np.sum(yt[:,1]*agp)+self.low_frequency_abduction[kiter]/lg
        h0i = self.up_height[kiter]
        h3i = self.down_height[kiter]
        Hvec = np.array([ h0i, h1i, h2i, h3i ])
        self.height_vector = Hvec

        ### Glottis shape
        x0 = xvec[0]
        x1 = xvec[1]
        x2 = xvec[2]
        dgl = xvec[3]-xvec[0]
        h0 = Hvec[0]
        h1i = Hvec[1]
        h2i = Hvec[2]
        if selfmod == 'smooth':
            A0 = (h1i-h0)/(x1-x0) # slope, part 1
            A1 = (h2i-h1i)/(x2-x1) # slope, part 2

            #### Separation point
            if (h1i>0) and (h2i>0): # No contact between masses
                if (h1i!=h2i): # if h1 different than h2
                    hsi = sep*h1i
                else:
                    hsi = h2i # separation height is h2
                if (h1i>=h2i) or (hsi>=h2i): #
                    hsi = h2i
            else:
                hsi = armin/lg   # if contact, separation height is 0 (set to eps to avoid singularities)

            Ag22 = 1/(hsi*lg+aChink)**2
            Rg2 = 0.5*rho/(lg**2)*(1/(hsi**2)-1/(h0**2))  # Bernoulli

            if (h1i!=h2i): # if plate 2 is non-horizontal
                Rg = 0.5*12*co_mu/lg*(1/A0*(1/(h0**2)-1/(h1i**2)) + 1/A1*(1/(h1i**2)-1/(hsi**2))) # viscous term
                Lg = rho/lg*(1/A0*np.log(h1i/h0)+1/A1*np.log(hsi/h1i)) # inertia
            else:  # if plate 2 is horizontal
                Rg = 12*co_mu/lg*(0.5/A0*(1/(h0**2)-1/(h1i**2)) + (x2-x1)/(h1i**3)) # viscous term
                Lg = rho/lg*(1/A0*np.log(h1i/h0)+(x2-x1)/h1i) # inertia
        elif  selfmod=='ishi':
            ag1 = h1i*lg
            ag2 = h2i*lg
            Rk1 = 0.19*rho/(ag1)**2
            Rk2 = (0.5 - ag2/ain*(1-ag2/ain)) / (ag2**2)*rho
            Lg1 = rho*(x1-x0)/ag1
            Lg2 = rho*(x2-x1)/ag2
            Rv1 = 12*co_mu*lg**2*(x1-x0)/(ag1**3)
            Rv2 = 12*co_mu*lg**2*(x2-x1)/(ag2**3)
            Rg2 = Rk1+Rk2
            Rg = np.array([Rv1, Rv2])
            Lg = np.array([Lg1,Lg2])
            hsi = min(h1i,h2i)
            hsi = max(hsi,armin/lg)#
            Ag22 = 1/(hsi*lg+aChink)**2

        if not Const.unst:
            Lg = Lg*0
        if Const.model =='smooth':
            Rg = np.array([Rg, Rg]/2)
            Lg = np.array([Lg, Lg]/2)

        self.mass_position = yt * agp
        if type(self) is Glottis:            
            Ac2 = 1/ac**2
            rv_sum = (12*lg*dgl*Ag22*np.sqrt(Ag22)+8*np.pi*lc*Ac2)*co_mu
            # rv_sum = 12*(dgl*Ag22+lc.*Ac2)*co_mu;
            rk_sum = 1.38*rho*(Ac2+Ag22)
            # rv_sum = Rg+8*pi*lc.*Ac2*co_mu;
            # rk_sum = 1.38*rho*Ac2+Rg2;
            Udc = (-rv_sum+np.sqrt(rv_sum*rv_sum+4*rk_sum*self.upstream_pressure))/(2*rk_sum)
    #        if ( isempty(Udc) or isnan(Udc) ); Udc = 0; end
        #     self.DC_flow = Udc;

            self.DC_flow = self.DC_flow+(Udc-self.DC_flow)*2*np.pi*500*T
            Udc = self.DC_flow
            self.area = np.max((armin, abs(hsi*lg)+aChink))

        self.bernoulli = Rg2
        self.resistance = Rg
        self.inductance = Lg
        self.separation_height = hsi

        return  Rg2, Rg, Lg, hsi, Udc

    def updateparam( self, Const ):
        # Adapt the values of the mass and stiffness of vocal folds to match the
        # fundamental frequency fo

        kiter = Const.k_iter
        fo = self.fundamental_frequency[kiter]
        dab = max(0.1,1-self.partial_abduction[kiter])
        mvfi = self.initial_mass

        if (np.isinf(fo) or np.isnan(fo) or fo <= 1 ):
            fo = 1

        mvfo = mvfi*dab
        kvfo = 4*np.pi**2*fo**2*mvfo
        kco = np.array([ kvfo[0,0], kvfo[1,0] ])*self.Qrc
        rvfo = self.Qr*np.sqrt(kvfo*mvfo/2)

        #### CHECK for contact
        kctc = self.contact_stiffness
        rctc = self.contact_damping
        self.isContact = False

        if (self.height_vector[1] <= Const.blim):  # contact at the 1st mass location
            kvfo[:,0] = kctc*kvfo[:,0] # New stiffness for 1st mass
            rvfo[:,0] = rctc*np.sqrt(kvfo[:,0]*mvfo[:,0]/2)#rvf(:,1)+2*sqrt(kvf(:,1).*mvf(:,1)); # New damping for 1st mass
            self.isContact = True

        if (self.height_vector[2] <= Const.blim): # contact at the 2nd mass location
            kvfo[:,1] = kctc*kvfo[:,1] # New stiffness for 2nd mass
            rvfo[:,1] = rctc*np.sqrt(kvfo[:,1]*mvfo[:,1]/2) #rvf(:,2)+2*sqrt(kvf(:,2).*mvf(:,2)); # New damping for 2nd mass
            self.isContact = True

        self.mass = mvfo
        self.stiffness = kvfo
        self.coupling_stiffness = kco
        self.damping = rvfo

        return mvfo, kvfo, rvfo, kco

    def updatemassishiflan(self, Const):

        ### Glottis shape
        rho = Const.rho
        T = Const.T

        etha_k1 = self.etha_k1
        etha_k2 = self.etha_k2
        etha_h1 = self.etha_h1
        etha_h2 = self.etha_h2
        xvec = self.x_position
        lg = self.length

        ytm1 = self.mass_position_nm1
        ytm2 = self.mass_position_nm2
        ago = self.low_frequency_abduction[Const.k_iter]
        U1 = self.inst_flow
        m1 = self.mass[1,0]
        m2 = self.mass[1,1]
        k1 = self.stiffness[1,0]
        k2 = self.stiffness[1,1]
        kc = self.coupling_stiffness[0]
        Ps = self.upstream_pressure

        x0 = xvec[0]
        x1 = xvec[1]
        x2 = xvec[2] #x3 = xvec(4)
        d1 = x1-x0
        d2 = x2-x1
        # h0 = hvec[0] h1i = hvec(2) h2i = hvec(3) h3 = hvec(4)

        y1m1 = np.sum(ytm1[:,0])
        y2m1 = np.sum(ytm1[:,1])
        y1m2 = np.sum(ytm2[:,0])
        y2m2 = np.sum(ytm2[:,1])
        ag1m1 = y1m1*lg + ago
        ag2m1 = y2m1*lg + ago
        ago1 = ago
        ago2 = ago
        h1 = 3*k1
        h2 = 3*k2

        # Pressure acting on m1
        Pm1 =  Ps - 1.37*rho/2*(U1/(ag1m1))**2 \
        - 0.5*( self.resistance[0]*U1+2*self.inductance[0]/T*U1-self.Qu1 )
        self.Qu1 = 4/T*self.inductance[0]*U1-self.Qu1

        F1 = d1*lg*Pm1 # Force due to Pm1

        # Pressure acting on m2
        F2 = d2*lg*( Pm1 - 0.5*(np.sum(self.resistance))*U1
                    -2*(np.sum(self.inductance))/T*U1+self.Qu2
                    -rho/2*U1**2*( 1/ag2m1**2-1/ag1m1**2 ) )
        self.Qu2 = 4/T*(np.sum(self.inductance))*U1-self.Qu2

        # Damping ratio
        r1 = 2*0.1*np.sqrt(m1*k1)
        r2 = 2*0.6*np.sqrt(m2*k2)

        # Computation of every possible case.
        #1 and 2 open
        y1oo = ( m1/T**2*(2*y1m1-y1m2) + \
            r1/T*y1m1-k1*etha_k1*y1m1**3 - \
            kc*(y1m1-y2m1) + F1) \
            /( m1/T**2+r1/T+k1 )
        y2oo = ( m2/T**2*(2*y2m1-y2m2) + \
            r2/T*y2m1-k2*etha_k2*y2m1**3 - \
            kc*(y2m1-y1m1) + F2) \
            /( m2/T**2+r2/T+k2 )

        # 1 open and 2 closed
        r2 = 2*1.6*np.sqrt(m2*k2)
        y1of = ( m1/T**2*(2*y1m1-y1m2) + \
            r1/T*y1m1-k1*etha_k1*y1m1**3 -\
            kc*(y1m1-y2m1) +Ps*d1*lg)\
            /( m1/T**2+r1/T+k1 )
        y2of = ( m2/T**2*(2*y2m1-y2m2) + \
            r2/T*y2m1-k2*etha_k2*y2m1**3 -\
            kc*(y2m1-y1m1) + Ps*d2*lg - \
            h2*(ago2*0.5/lg+etha_h2*(y2m1+ago2*0.5/lg)**3) )\
            /( m2/T**2+r2/T+k2+h2)

        # 1 closed and 2 open
        r1 = 2*1.1*np.sqrt(m1*k1)
        r2 = 2*0.6*np.sqrt(m2*k2)
        y1fo = ( m1/T**2*(2*y1m1-y1m2) + \
            r1/T*y1m1-k1*etha_k1*y1m1**3 -\
            kc*(y1m1-y2m1) + Ps*d1*lg - \
            h1*(ago1*0.5/lg+etha_h1*(y1m1+ago1*0.5/lg)**3))\
            /( m1/T**2+r1/T+k1+h1 )
        y2fo = ( m2/T**2*(2*y2m1-y2m2) + \
            r2/T*y2m1-k2*etha_k2*y2m1**3 -\
            kc*(y2m1-y1m1) + 0)\
            /( m2/T**2+r2/T+k2 )

        # 1 and 2 closed
        r2 = 2*1.6*np.sqrt(m2*k2)
        y1ff = ( m1/T**2*(2*y1m1-y1m2) + \
            r1/T*y1m1-k1*etha_k1*y1m1**3 -\
            kc*(y1m1-y2m1) + Ps*d1*lg - \
            h1*(ago1*0.5/lg+etha_h1*(y1m1+ago1*0.5/lg)**3))\
            /( m1/T**2+r1/T+k1+h1 )
        y2ff = ( m2/T**2*(2*y2m1-y2m2) + \
            r2/T*y2m1-k2*etha_k2*y2m1**3 -\
            kc*(y2m1-y1m1) + 0 - \
            h2*(ago2*0.5/lg+etha_h2*(y2m1+ago2*0.5/lg)**3) )\
            /( m2/T**2+r2/T+k2+h2)

        if y1oo > -ago1/(2*lg):
            if y2oo > -ago2/(2*lg):
                y1 = y1oo
                y2 = y2oo
            else:
                y1 = y1of
                y2 = y2of
        else:
            if y2fo > -ago2/(2*lg):
                #           1 closed 2 open
                y1 = y1fo
                y2 = y2fo
            else:
                #           1 closed 2 closed
                y1 = y1ff
                y2 = y2ff

        ytm2 = ytm1
        yt = np.hstack((y1/2*np.ones((2,1)), y2/2*np.ones((2,1))))
        ytm1 = yt

        self.mass_position_nm2 = ytm2
        self.mass_position_nm1 = ytm1
        self.mass_position = yt
        self.applied_force = np.array([F1, F2])

        return yt, ytm1, ytm2

    def updatemassposition(self, Const):
        
        Fvf = self.applied_force
        fs = 1/Const.T
        kc = self.coupling_stiffness
        mvf = self.mass
        ke = self.stiffness
        re = self.damping
        ytm1 = self.mass_position_nm1
        ytm2 = self.mass_position_nm2
        yt = np.zeros((2,2))
        # activation = self.activation[Const.k_iter]
        activation = 1

        F1u = np.sum(Fvf[:,0])
        F2u = np.sum(Fvf[:,1])
        F1d = F1u
        F2d = F2u
        kcup = kc[0]
        kcdown = kc[1]

        tmp1_A = mvf[0,0]*fs**2+ke[0,0]+re[0,0]*fs+kc[0]
        tmp1_B = mvf[0,0]*fs**2*(2*ytm1[0,0]-ytm2[0,0])+re[0,0]*fs*ytm1[0,0]
        tmp2_A = mvf[0,1]*fs**2+ke[0,1]+re[0,1]*fs+kc[0]
        tmp2_B = mvf[0,1]*fs**2*(2*ytm1[0,1]-ytm2[0,1])+re[0,1]*fs*ytm1[0,1]

        yt[0,0] = (F1u+tmp1_B+(kcup/tmp2_A)*(F2u+tmp2_B))/(tmp1_A-(kcup**2)/tmp2_A)
        yt[0,1] = (F2u+tmp2_B+(kcup/tmp1_A)*(F1u+tmp1_B))/(tmp2_A-(kcup**2)/tmp1_A)

        tmp1_A = mvf[1,0]*fs**2+ke[1,0]+re[1,0]*fs+kc[1]
        tmp1_B = mvf[1,0]*fs**2*(2*ytm1[1,0]-ytm2[1,0])+re[1,0]*fs*ytm1[1,0]
        tmp2_A = mvf[1,1]*fs**2+ke[1,1]+re[1,1]*fs+kc[1]
        tmp2_B = mvf[1,1]*fs**2*(2*ytm1[1,1]-ytm2[1,1])+re[1,1]*fs*ytm1[1,1]

        yt[1,0] = (F1d+tmp1_B+(kcdown/tmp2_A)*(F2d+tmp2_B))/(tmp1_A-(kcdown**2)/tmp2_A)
        yt[1,1] = (F2d+tmp2_B+(kcdown/tmp1_A)*(F1d+tmp1_B))/(tmp2_A-(kcdown**2)/tmp1_A)
        ytm2 = ytm1
        ytm1 = yt

        self.mass_position_nm2 = ytm2
        self.mass_position_nm1 = ytm1
        self.mass_position = yt*activation

        return yt, ytm1, ytm2

    def appliedforce(self, Const):
        # Compute the applied force on vocal folds

        sep = Const.sep
        T = Const.T

        xvec = self.x_position
        Hvec = self.height_vector
        (x0, x1, x2, x3) = xvec
        (h0, h1i, h2i, h3) = Hvec

        Ps = self.upstream_pressure
        Psup = self.downstream_pressure
        DeltaPi = Ps-Psup
        Ug = self.inst_flow
        Ugim1 = self.flow_nm1
        lg = self.length

        A = [(h1i-h0)/(x1-x0),
             (h2i-h1i)/(x2-x1),
             (h3-h2i)/(x3-x2)]
        B = [h0-A[0]*x0,
             h1i-A[1]*x1,
             h2i-A[2]*x2]

        if (h1i>0) and (h2i>0): # No contact between masses
            if (h1i!=h2i): # if h1 different than h2
                 hsi = sep*h1i
                 xsi = (hsi-h1i)/A[1]+x1
            else:
                 xsi = x2 # if h1 = h2
                 hsi = h2i # separation height is h2
            if (h1i>=h2i) or (hsi>=h2i): #
                hsi = h2i
                xsi = x2   # separation point is x2
        else:
            hsi = 1e-11   # if contact, separation is 0
            xsi = 1e-11   #

        if DeltaPi>=0:
            dUg_tmp = (Ug-Ugim1)/T
            Fvf = forcecompute(Ug,lg,dUg_tmp,Ps,Psup,xvec,h1i,h2i,A,B,h0,xsi,hsi, Const)
        else:
            dUg_tmp = (-Ug-Ugim1)/T
            Fvf = forcecompute(Ug,lg,dUg_tmp,Psup-Ps,0,xvec,h2i,h1i,A,B,h0,xsi,hsi, Const)

        Ugim1 = Ug
        self.flow_nm1 = Ug
        self.applied_force = Fvf

        return Fvf, Ugim1


class Glottis(Oscillator):
#GLOTTIS Summary of this class goes here
#   Detailed explanation goes here

    # Input
    subglottal_pressure = None

    # Output
    #### Instantaneous
    DC_flow = 0
    max_opening = 4e-5
    area = None

    #### Time evolution
    DC_vector = None

      # Constructor method
    def __init__(self, **kwargs):
        isArg = False
        for key, value in kwargs.items():
            setattr(self, key, value)
            isArg = True
        if not isArg:

            Qam = 1
            Qak = 1
            Qam2 = 0.2
            Qak2 = 0.2

            rho_folds = 1020
            hfolds = 1e-2*(0.065+0.05+0.115)#3e-3
            kspring = 10

            self.rest_position = 5e-5*np.ones((2,2))
            self.mass_position = 5e-5*np.ones((2,2))
            self.width = 0.003
            x1 = self.width/(1+Qam2)  # 1st mass
            x2 = self.width   # 2nd mass
            x3 = x2+x1*1e-11  # end of constriction
            self.x_position = np.array([ 0, x1, x2, x3 ])
            self.length = 0.016

            self.Qr = 0.2
            self.Qrc = 0.5

            # Mechanical parameters
            m1ui = rho_folds*self.length*self.width*hfolds/(1+Qam2)/2*1.2 # upper mass 1 (kg)
            k1ui = kspring # upper spring stiffness 1 (N/m)
            r1ui = self.Qr*np.sqrt(k1ui*m1ui/2) # damping of the upper mass 1
            m2ui = m1ui*Qam2 # upper mass 2 (kg)
            k2ui = k1ui*Qak2 # upper spring stiffness 2 (N/m)
            kcui = k1ui*self.Qrc # upper coupling spring (N/m)
            r2ui = self.Qr*np.sqrt(k2ui*m2ui/2) # damping of upper mass 2
            m1di = m1ui*Qam  # lower mass 1 (kg)
            k1di = k1ui*Qak  # lower spring stiffness 1 (N/m)
            m2di = m2ui*Qam  # lower mass 2 (kg)
            r1di = self.Qr*np.sqrt(k1di*m1di/2) # damping of lower mass 1
            k2di = k2ui*Qak # lower spring stiffness 2 (N/m)
            kcdi = kcui*Qak  # lower coupling stiffness (N/m)
            r2di = self.Qr*np.sqrt(k2di*m2di/2) # damping of lower mass 2

            self.initial_mass = np.array([ [m1ui, m2ui], [ m1di, m2di] ]) # mass matrix
            self.initial_stiffness = np.array([ [ k1ui, k2ui], [ k1di, k2di ] ])# stiffness matrix
            self.initial_damping = np.array([ [ r1ui, r2ui], [ r1di, r2di ] ]) # damping matrix
            self.initial_coupling_stiffness = np.array([ kcui, kcdi ]) # coupling stiffness

            self.etha_k1 = 100
            self.etha_k2 = 100
            self.etha_h1 = 500
            self.etha_h2 = 500
            self.contact_stiffness = 4
            self.contact_damping = 2.2
            self.upstream_pressure = 0
            self.downstream_pressure = 0

class Tongue(Oscillator):
    # Tongue Summary of this class goes here
    #   Detailed explanation goes here


    position = None
    oscillation_factor = None
    Qdown = 0
    Qup = 0
    waveguide = None
    num_oscillators = None
    contact_stiffness = 1.5
    contact_damping = 1.2

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def updatepressure(self,  objwvg, Const):

        T = Const.T
        ptt = self.position
        utt = objwvg.flow_time[ptt]
        uttp1 = objwvg.flow_time[ptt+1]
        self.flow_inst = objwvg.flow_time[ptt-1]
        self.upstream_pressure = objwvg.pressure_time[ptt-1] - (objwvg.Rj[ptt-1]+2/T*objwvg.Lj[ptt-1]+objwvg.Rcp[ptt-1])*utt+self.Qup
        self.downstream_pressure = objwvg.pressure_time[ptt+1] - (objwvg.Rj[ptt+1]+2/T*objwvg.Lj[ptt+1]+objwvg.Rcp[ptt+1])*utt-self.Qdown

        self.Qup = 4/T*objwvg.Lj[ptt-1]*utt-self.Qup
        self.Qdown = 4/T*objwvg.Lj[ptt+1]*uttp1-self.Qdown




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Oct 26 13:56:52 2020

@author: benjamin
"""

import numpy as np
from .utils import (nextpow2, symmetricifft, cepstrum, buffer,
                    flatten_ceps, flatten_ceps_matrix, optimize_cepstrum, 
                    get_order_from_sig)
import scipy.signal as ssg

def change_cepstrum(x2, Spx2, x_out, nwin, nhop, nfft, txx, isFlat, isNorm,
                    Sxx, Sxx_target):
    
    xwin_fft, xx_win = overlap_fft(x2, nwin, nhop, nfft)
    if isFlat:
        CepOrder = get_order_vec(Spx2, txx)
        xwin_fft = flatten_ceps_matrix(xwin_fft, xx_win, 
                                           CepOrder, Spx2.sampling_frequency)
        Sxx = 1
    xwin_fft = xwin_fft * np.abs(Sxx_target) / np.abs(Sxx)
    xnew = symmetricifft(xwin_fft, nfft)
    x_out = overlap_add(xnew, xx_win, x_out, nhop, isNorm)
    x_out[np.isnan(x_out)] = 0
    return x_out

def ola_param(nwin, novlp=None, sr=None):
    
    if novlp is None:
            novlp = int(nwin * 3 / 4)
    novlp = int(novlp)
    if sr is None:    
        nfft = int(nwin)
    else:
        nfft = 2**nextpow2(sr)
        
    nhop = nwin - novlp
    return novlp, nfft, nhop

def init_ola(x, nwin, nhop):
    
    zeros2add = np.zeros((nhop, 1))
    x2 = np.vstack((zeros2add, x.reshape(-1, 1), zeros2add)).squeeze()
    x_out = np.zeros_like(x2)
       
    return x2, x_out

def get_order_vec(sp, tx):
    
    pitch = sp.getpitch()
    pitch.timeinterpolate(tx)
    f0 = pitch.values
    f0[np.isnan(f0)] = 125
    f0[f0<=0] = 125
    return 2 * np.round(sp.sampling_frequency / f0)

def check_index(x, k, nwin, nhop):
    
    w = ssg.hann(nwin)
    deb = int(k * nhop)
    fin = deb + nwin - 1
    isC = True
    if fin > len(x) - 1:
        fin = len(x) - 1
        isC = False
    idx = np.arange(deb, fin+1).astype(int)
    xwin = x[idx] * w[:len(idx)]
    
    return xwin, idx, isC

def spec_frame(x, k, nwin, nhop, nfft):
    
    L = int(nfft / 2 + 1)
    xwin, idx, isC = check_index(x, k, nwin, nhop)
    fft_xwin = np.fft.fft(xwin, nfft)[:L]
    
    return xwin, fft_xwin, isC, idx

def change_frame(fft_xwin, h_new, h_orig, nfft, idx, x_out, xwin, 
                 isChange=True, isNorm=True, sr=8000):
    
    init_en = sum(abs(xwin)**2)
    if init_en != 0 or not np.isnan(init_en):        
        if isChange:
            order = get_order_from_sig(xwin, sr)
            fft_xwin, Cx_flat = flatten_ceps(xwin, fft_xwin, sr, order, tol=1e-4)
            # fft_new = optimize_cepstrum(xwin, fft_xwin, h_new, Cx_flat, sr,
            #   order, max_iter=10)
            fft_new = fft_xwin * abs(h_new) #  / abs(h_orig)
            x_new = symmetricifft(fft_new, nfft)
        else: 
            x_new = xwin
        
        x2add = x_new[:len(idx)]    
        new_en = sum(abs(x2add)**2)
        if new_en == 0 or np.isnan(new_en):            
            pass
        else:
            if isNorm:
                x_out[idx] = x_out[idx] + x2add * init_en / new_en
            else:
                x_out[idx] = x_out[idx] + x2add  
    
    return x_out

def check_frame(k, isC, maxFrame):
    
    k += 1
    if k >= maxFrame:
        isC = False
        
    return k, isC

def overlap_add(x_transformed, x_initial, x_out, nhop, isNorm=True):
    
    new_en = np.sum(np.abs(x_transformed**2), 0)    
    nb_win, nb_frame = x_transformed.shape
    
    if isNorm:
        init_en = np.sum(np.abs(x_initial**2), 0)
        init_en[new_en == 0] = 1
        new_en[new_en == 0] = 1
        ratio_en = init_en / new_en
        ratio_en[np.isnan(ratio_en)] = 1
    else:
        ratio_en = np.ones(nb_frame)
    
    idx1 = 0
    for k, x_add in enumerate(x_transformed.T):
        
        idx2 = idx1 + nb_win
        if idx2 >= len(x_out):
            idx2 = len(x_out)
        x_out[idx1:idx2] += x_add[:idx2-idx1] * ratio_en[k]
        idx1 += nhop
        
    return x_out[nhop:-nhop] / 1.5

def overlap_fft(x, nwin, nhop, nfft):
    
    xx = buffer(x, nwin, nhop)
    L = int(nfft/2+1)
    w = ssg.hann(nwin)
    xx_win = xx * w.reshape(-1,1)
    xwin_fft = np.fft.fft(xx_win, nfft, axis=0)[:L, :]
    return xwin_fft, xx_win


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Sep 13 09:28:58 2018

@author: benjamin
"""
import numpy as np
import scipy
from ._set import miscstruct
from .utils import nextpow2

def initparam(obj, param):
    
    if param is None:
        param = miscstruct()        
  
    if not hasattr(param, 'win_length_sec'):
        param.win_length_sec = 0.04
    if not hasattr(param, 'win_length'):
        param.win_length = 2**nextpow2(param.win_length_sec * obj.sampling_frequency)
    if not hasattr(param, 'hop_factor'):
        param.hop_factor = 4
    if not hasattr(param, 'threshold'):
        param.threshold = 3.5
    if not hasattr(param, 'nfft'):
        param.nfft = 2**nextpow2(param.win_length * 10)
    if not hasattr(param, 'lfmed'):
        param.lfmed = 250
    if not hasattr(param, 'num_period'):
        param.num_period = 4
    if not hasattr(param, 'fomin'):
        param.fomin = 80
    if not hasattr(param, 'fomax'):
        param.fomax = 300
    if not hasattr(param, 'num_harm'):
        param.num_harm = np.floor(obj.sampling_frequency / 2 / param.fomin)
    if not hasattr(param, 'partial_length'):
        param.partial_length = 7
    if not hasattr(param, 'search_range'):
        param.search_range = 7
    if not hasattr(param, 'num_partial'):
        param.num_partial = 6
    if not hasattr(param, 'num_zero'):
        param.num_zero = 1
    if not hasattr(param, 'pct'):
        param.pct = 90
    if not hasattr(param, 'amplitude_estimator'):
        param.amplitude_estimator = 'lse'
    if not hasattr(param, 'std_estimator'):
        param.std_estimator = 'std'
    if param.amplitude_estimator.lower() != 'lse' and param.amplitude_estimator.lower() != 'wlse':
        print('Warning: the amplitude estimator is badly defined. It has been set to LSE (default)')
        param.amplitude_estimator = 'lse'
    if not hasattr(param, 'jitt'):
        param.jitt = False
    if not hasattr(param, 'shim'):
        param.shim = False
    if (param.jitt or param.shim) and (not hasattr(param, 'ord')):
        param.ord = np.ceil(np.min([param.win_length / 3, 501]))
    if param.jitt or param.shim:
        param.fcoef = scipy.signal.firwin(int(param.ord), 
                                 float(2 * param.partial_length / (obj.sampling_frequency / 2)))
    if (param.jitt or param.shim) and (not hasattr(param, 'methfreq')):
        param.methfreq = 'zcr'
    if not hasattr(param, 'verbosity'):
        param.verbosity = False      
        
    return param

def findpartials(x, fs, win, param):
    """ returns a cleaned version of the periodogram
    keeps interval of the periodogram above peaks that are above the noise
    floor (estimated by a median filtering of the periodogram) """

    Nfft = int(param.nfft)
    f = fs*np.arange(0,Nfft)/Nfft
    P = abs(np.fft.fft(x*win,Nfft))**2
    P = P/sum(P)

    nw = param.lfmed*np.ceil(Nfft/param.nfft)
    if np.mod(nw,2)==False:
        nw += 1
    if param.pct > 0:
        pth = scipy.ndimage.filters.percentile_filter(P, int(param.pct), int(nw))        
        P[P < pth] = 0
    
    P[f < param.fomin] = 0
    pks = np.zeros(len(P))
    m = np.logical_and((P[1:-1] > P[2:len(P)]),(P[1:-1] > P[0:-2]))
    locs = np.argwhere(m)+1 # find peaks
    prng = np.ceil(param.partial_length*fs/Nfft)

    # keep the periodogram around the peaks above the noise floor
    for ki in range(0,len(locs)):
        if (locs[ki] > prng) and (locs[ki]+prng < len(pks)):
            partial_range = np.arange(locs[ki]-prng, locs[ki]+prng+1)
            pks[partial_range.astype(int)] = P[partial_range.astype(int)]
            
    return pks

def findf0(pks, fs, param):
    """ Estimation of the fundamental frequency of the windowed signal S

    Input arguments:
    
    pks: cleaned periodogram
    fs: sampling frequency (in Hz)
    param : structure containing various parameters """

    f = np.arange(0,param.nfft)/ param.nfft*fs
    idx1 = np.argwhere(f >=param.fomin)[0,0]
    idx2 = np.argwhere(f <= param.fomax)[-1,-1]
     
    ## cumulated periodogram
    cum_per = np.zeros(len(pks))

    Npartials = param.num_partial

    for ki in range(idx1,idx2+1):#idx1:idx2
        for kj in range(1,min((np.floor(len(pks)/ki), Npartials))+1):
            cum_per[ki] = cum_per[ki] + pks[ki * kj]
            
    # keep only if fundamental is present
    cum_per[pks == 0] = 0
    # first estimation of f0
    maxcumul = max(cum_per)
    if maxcumul == 0:
        return 0
    idxf0 = np.argmax(cum_per)
    # find values of cum_per close to the maximum
    idxfindf0 = np.argwhere(cum_per > 0.95 * maxcumul)

    idxf0 = max(idxfindf0[np.mod(idxfindf0,idxf0)==0])
    # solve the problem when f0/2 occurs in the cumulated periodog

    if idxf0 > np.ceil(param.search_range*fs/param.nfft):
        idx1 = int(idxf0-np.ceil(param.search_range*fs/param.nfft))
        idx2 = int(idxf0+np.ceil(param.search_range*fs/param.nfft)+1)
        maxpks = max(pks[idx1:idx2])
        idxf = np.argmax(pks[idx1:idx2])
    else:
        idx2 = int(idxf0+np.ceil(param.search_range*fs/param.nfft)+1)
        maxpks = max(pks[0:idx2])
        idxf = np.argmax(pks[0:idx2])

    if ((not maxpks) or (maxpks/sum(pks) < 1e-3)):# || (np.abs(param.search_range+1-idxf) < param.partial_lengthh))
        return 0  # If empty, no voicing, the signal contains only noise

    idxf0 = int(idxf0+idxf-np.ceil(param.search_range*fs/param.nfft)-1)

    if not idxf0:
        return 0

    return f[idxf0]#, cum_per[idxf0]

def f0reeval(f0, x, fs, param):
    """ correction of f0/2 and f4/2 using the derivative criterion """
    for ki in range(0,2):

        derrive = np.zeros(len(f0))
        derrive[1:len(derrive)] = np.diff(f0)
        derrive[f0==0]=0
        f0_ = np.zeros(len(f0))
        f0_[1:len(f0_)]= f0[0:-1]
        derrive[f0_ == 0]=0
        detection = derrive/(f0+1e-5)*100
        f0[detection <-50] = 0
        f0 = scipy.signal.medfilt(f0,3)
        f0[np.logical_or(f0 < param.fomin, f0 > param.fomax)] = 0
    fk = np.zeros((len(f0), int(param.num_harm)))
 
    kit = param.hop+np.ceil(param.win_length/2)-1
    param2 = param   
    param2.pct = -1

    for cmpt in range (0,len(f0)):
        if f0[cmpt] >= param.fomin:
            tow = np.floor(1./f0[cmpt]*fs)
            npts = np.ceil(param.win_length/tow)
            if abs((npts-1)*tow-param.win_length) < abs(npts*tow-param.win_length):
                npts -= 1
            nw = npts*tow
            iw = np.arange(max((0,kit-np.floor(nw/2))),min((kit+np.floor(nw/2)-1+np.mod(nw,2)+1,len(x))))
            xw = x[iw.astype(int)]
            w = np.hanning(len(xw)+1)
            w = w[1:len(w)]
#             param2.lfmed = ceil(f0(cmpt)*1.5);            
            pks = findpartials(xw, fs, w, param2)
            fk[cmpt,:] = relocatePartials( pks, xw, fs, f0[cmpt], param2)   
        kit += param.hop; 
    return fk

def relocatePartials( pks, x, fs, f0, param ):

    nfft = int(param.nfft)
    fkm = np.zeros(int(param.num_harm))   
    w = np.hanning(len(x)+1)
    w = w[1:len(w)]
    f = np.arange(0,nfft)/ nfft*fs

    idxf0 = int(np.argwhere(f >=f0 )[0])
    fkm[0] = f0
    idxfi = idxf0
    
    yg = np.fft.fft(x*w, nfft)
    L = int(np.ceil(nfft/2)+1)
    yg = yg[0:L]
    f = f[0:L]
    pks = pks[0:L]
    nzrow = 0
    prng = np.ceil(param.search_range*fs/nfft)
    
    for ki in range(1,int(param.num_harm)+1):
        idxfi += idxf0
    #     idxfi=ki*idxf0;#+idxfi;
        if idxfi<=len(pks):
            idx1 = int(max(idxfi-prng,0))
            idx2 = int(min(idxfi+prng+1,len(pks)))
            valeur = np.max(abs(pks[idx1:idx2]))
            idxf = np.argmax(abs(pks[idx1:idx2]))
            if valeur != 0:
                idxfi += idxf-prng-1
                idxtt = int(idxfi)
                idx1 = int(max(idxfi-prng,0))
                idx2 = int(min(idxfi+prng+1,len(yg)))
                idxf = np.argmax(abs(yg[idx1:idx2]))
                idxfi2 = int(idxfi+idxf-prng-1)
                ir = np.arange(max(idxfi2-2,0),min(idxfi2+3,len(yg)))
                if max(ir)<=len(f):
                    wos = f[ir.astype(int)]
                    los = np.log(abs(yg[ir.astype(int)]))
                    p = np.polyfit(wos, los,2)
                    fkm[ki] = -p[1]/(2*p[0])
                    if (fkm[ki] < 0) or (fkm[ki] > fs/2) or (abs(fkm[ki]-f[int(idxfi2)]) > prng):
                        fkm[ki] = f[idxtt]
                        idxfi = int(idxtt)
                else:
                    fkm[ki] = ki*f0
        else:
            break        
        if np.isnan(fkm[ki]):
            fkm[ki] = 0
        if fkm[ki]:
            nzrow = 0;
        else:
            nzrow += 1
            if param.num_zero and nzrow >= param.num_zero:
                return fkm
            
    return fkm

def instprop( x, fs=1):
    
    """ returns instantaneous lower and upper amplitude """ 
    xm = np.mean(x)
    xmm = x-xm
    
    x_anal = scipy.signal.hilbert(xmm)
    phix = np.unwrap(np.angle(x_anal))    
#    ifq = (np.diff(phix)/(2.0*np.pi)*fs)
    ifq = 1./(2*np.pi)*(np.gradient(phix,1./fs))
    upenv = abs(x_anal)
    
    x = -x;
    xm = np.mean(x); 
    xmm = x-xm;
    x_anal = scipy.signal.hilbert(xmm)
    lowenv = abs(x_anal)
    
    return ifq, upenv, lowenv

def instfreq(x, fs, fo):
    """ returns instantaneous frequency """
    
    tow = np.floor(1/fo*fs)
    x = x-np.mean(x)
    nx = len(x)
    txi = np.arange(0,nx-0.99,0.01)
    f1 = scipy.interpolate.interp1d(np.arange(0,nx),x)
#    foint = f1(tfo)
    xi = f1(txi)
    px = xi[0:-1]*xi[1:len(xi)]
#    zc = np.argwhere(px <= 0)  
    zc = np.argwhere(px <= 0)
    zc = zc.reshape(zc.size)
    zct = txi[zc[0:len(zc):2]]   
    foi = np.zeros(len(x))
    
    indx = scipy.signal.argelrextrema(xi, np.greater)
    hmax = xi[indx]
    imin = scipy.signal.argelrextrema(xi, np.less)
    hmin = xi[imin]
    lmax = len(hmax)
    lmin = len(hmin);
    lm = min((lmax,lmin))
    hmax = hmax[0:lm] 
    hmin = hmin[0:lm]
    
    fo = fs/np.diff(zct)
    perloc = 0.5*(zct[0:-1]+zct[1:len(zct)])
    
    if len(perloc) >= 2:
        if perloc[0] > 2*tow:
            fo = np.append(0,fo)# zeros(1).append(fo)
            perloc = np.append(1,perloc)#np.ones(1).append(perloc)            
        if nx-perloc[-1] > 2*tow:
            fo = np.append(fo,0)#.append(0)
            perloc = np.append(perloc,nx)#.append(nx)

#        f1 = scipy.interpolate.interp1d(perloc, fo, kind='cubic',bounds_error=False,fill_value=0)
        f1 = scipy.interpolate.PchipInterpolator(perloc,fo)
        foi = f1(txi)
        foi = foi[0:len(foi):100]
    
    return foi


# -*- coding: utf-8 -*-
"""
Created on Tue Jun 23 15:32:24 2020

@author: benjamin
"""

import numpy as np
import scipy.interpolate as scintp
from .utils import *
import spychhiker.utils as utils
import spychhiker.plottools as plottools
import copy
from .stats import RandomVar, ProsodyParam
from .prosody import Contour

class TimeFreq:

    parent = None
    time_frequency_matrix = None
    time_vector = None
    frequency_vector = None
    window = None
    overlap = None
    nfft = None

    def __init__(self, *args):
        nargin = len(args)
        for k in range(0, nargin, 2):
            key, value = args[k:k+2]
            setattr(self, key, value)

    def copy(self):
        return copy.deepcopy(self)

    def interpolate(self, vec_out, axis='time', method='linear'):

        y = self.time_frequency_matrix
        if axis == 'freq':
            vec_in = self.frequency_vector
            axis_num = 0
        elif axis == 'time':
            vec_in = self.time_vector
            axis_num = 1

        f1 = scintp.interp1d(vec_in, y, kind=method, axis=axis_num,
                             fill_value='extrapolate')
        y_out = f1(vec_out)

        self.time_frequency_matrix = y_out
        if axis == 'freq':
            self.frequency_vector = vec_out
        elif axis == 'time':
            self.time_vector = vec_out

    def tfplot(self, dyn=100, norm=False,
               pitch_plot=False, formant_plot=False):

        tvec = np.arange(len(self.parent.signal)) / self.parent.sampling_frequency
        lbl = self.parent.phonetic_labels
        inst = self.parent.phonetic_instants

        Pxx = abs(self.time_frequency_matrix)**2
        if norm:
            Pxx = Pxx / np.max(Pxx)

        fsp = self.frequency_vector
        tsp = self.time_vector
        Pxx = 10*np.log10(Pxx)
        signal = self.parent.signal

        pitch_time_points = []
        pitch_values = []
        formants_time = []
        formants_freq = []

        if pitch_plot:
            if self.parent.pitch is None:
                print('Warning: no pitch found!')
            else:
                pitch_time_points = self.parent.pitch.time_points
                pitch_values = self.parent.pitch.values

        if formant_plot:
            if self.parent.formants is None:
                print('Warning: no formants found!')
            else:
                formants_time = self.parent.formants.time_vector
                formants_freq = self.parent.formants.frequency.T

        return plottools.plot_spec(Pxx, tsp, fsp, dyn, tvec, signal,
              lbl, inst, pitch_plot, pitch_time_points, pitch_values,
              formant_plot, formants_time, formants_freq)

class FreqSig:

    parent = None
    values = None
    frequency_vector = None
    window = None
    nfft = None

    def __init__(self, *args):
        nargin = len(args)
        for k in range(0, nargin, 2):
            key, value = args[k:k+2]
            setattr(self, key, value)

    def copy(self):
        return copy.deepcopy(self)

    def freqplot(self, dyn=100, norm=False):

        Pxx = np.abs(self.values)**2
        if norm:
            Pxx = Pxx / np.max(Pxx)
        return plottools.freq_plot(self.frequency_vector, Pxx, dyn)

    def interpolate(self, freq_out, isdB=False, method='linear'):
        y = self.values
        freq_in = self.frequency_vector
        if isdB:
            y = 10*np.log10(np.abs(y))
        else:
            y = np.abs(y)

        f1 = scintp.interp1d(freq_in,y,kind=method,fill_value='extrapolate')
        y_out = f1(freq_out)

        if isdB:
            y_out = 10**(y_out / 20)

        self.values = y_out
        self.frequency_vector = freq_out


class Formants:
    frequency = None
    bandwidth = None
    damping = None
    amplitude = None
    isForm = None
    trajectory = None

    time_vector = None
    parent = None
    stats = None

    # Constructor method
    def __init__(self, *args):
        nargin = len(args)
        for k in range(0, nargin, 2):
            key, value = args[k:k+2]
            setattr(self, key, value)

    def copy(self):
        return copy.deepcopy(self)

    def addzeros(self, A):

        freq = self.frequency
        damp = self.damping
        ampl = self.amplitude
        bw = self.bandwidth
        nF, nFrame = freq.shape
        nZ, nF2ch = A.shape
        freqMat = np.vstack((freq[:nF2ch-1,:], np.ones((1,nFrame))))
        freqzero = A @ freqMat

        zeroObj = Formants('parent', self.parent,
        'frequency', freqzero,
        'damping', damp[:nZ,:],
        'bandwidth', bw[:nZ,:],
        'amplitude', ampl[:nZ,:],
        'isForm', False,
        'time_vector', self.time_vector)

        return zeroObj
    
    def getcontour(self):
        
        nb_form = self.frequency.shape[0]
        self.trajectory = []
        for k in range(nb_form):
            trj_tmp = Contour("time_points", self.time_vector, 
                              "values", self.frequency[k,:], 
                              "parent", self)
            self.trajectory.append(trj_tmp)
            
    def update_trajectory(self):
        
        if self.trajectory is not None:
            self.interpolate(self.trajectory[0].time_points)
            self.frequency = []
            for trj in self.trajectory:
                self.frequency.append(trj.values)
                self.time_vector = trj.time_points
            self.frequency = np.array(self.frequency)
    
    def transform(self, sr, *args):

        freq = self.frequency
        damp = self.damping
        tVec = self.time_vector
        nForm, nFrame = freq.shape
        z = np.exp(2 * 1j * np.pi * freq / sr) * np.exp(-damp / sr)
        z_module = np.abs(z)
        z_angle = np.exp(1j * np.angle(z))
        dampcoeff = np.array([1])
        freqcoeff = np.array([1])

        for k in range(0,len(args),2):
            field, argvalue =  args[k:k+2]
            if field.lower() == 'frequency':
                freqcoeff = argvalue
                if isinstance(freqcoeff, (list, tuple, np.ndarray)) is False:
                    freqcoeff = np.array([freqcoeff])
            elif field.lower() == 'damping':
                dampcoeff = argvalue
                if isinstance(dampcoeff, (list, tuple, np.ndarray)) is False:
                    dampcoeff = np.array([dampcoeff])
            else:
                raise ValueError('Atttribute to modify not recognized. Please choose either frequency or damping')

        if len(freqcoeff.shape) == 1:
            if len(dampcoeff.shape) == 1:
                z_new = (z_module**dampcoeff) * (z_angle**argvalue)
            else:
                nForm2ch = dampcoeff.shape[0]
                if nForm2ch > nForm:
                    nForm2ch = nForm
                    idx = list(range(nForm))
                    idx.append(dampcoeff.shape[1])
                    dampcoeff = dampcoeff[:nForm, idx]
                z_new = z
                for kF in range(nFrame):
                    f_pos = damp[:nForm2ch, kF]
                    damp_vec = np.concatenate((f_pos.reshape(-1),1))
                    new_damp = dampcoeff * damp_vec
                    ratio_damp = new_damp / f_pos
                    z_module_tmp = z_module[:nForm2ch, kF]
                    z_angle_tmp = z_angle[:nForm2ch, kF]
                    z_new[:nForm2ch, kF] = (z_module_tmp**ratio_damp) * (z_angle_tmp**freqcoeff)
        else:
            nForm2ch = freqcoeff.shape[0]
            if nForm2ch > nForm:
                nForm2ch = nForm
                idx = list(range(nForm))
                idx.append(freqcoeff.shape[1])
                freqcoeff = freqcoeff[:nForm, idx]

            if len(dampcoeff.shape) == 1:
                ratio_damp = dampcoeff
                isVec = False
            else:
                nDamp2ch = dampcoeff.shape[0]
                if nDamp2ch > nForm:
                    nDamp2ch = nForm
                    idx = list(range(nForm))
                    idx.append(dampcoeff.shape[1])
                    dampcoeff = dampcoeff[:nForm, idx]
                isVec = True

            z_new = z
            for kF in range(nFrame):
                f_pos = freq[:nForm2ch, kF]
                freq_vec = np.append(f_pos, 1)
                new_freq = np.squeeze(freqcoeff @ freq_vec.reshape(-1,1))

                ratio_freq = new_freq / f_pos
                z_module_tmp = z_module[:nForm2ch, kF]
                z_angle_tmp = z_angle[:nForm2ch, kF]
                if isVec:
                    damp_pos = damp[:nForm2ch, kF]
                    damp_vec = np.concatenate((f_pos.reshape(-1),1))
                    new_damp = dampcoeff * damp_vec
                    ratio_damp = new_damp / damp_pos
                z_new[:nForm2ch, kF] = (z_module_tmp**ratio_damp) * (z_angle_tmp**ratio_freq)

        freq_new = np.angle(z_new) / 2 / np.pi * sr
        damp_new = -np.log(abs(z_new)) * sr
        bw_new = damp_new / np.pi

        obj_out = Formants('frequency', freq_new,
                'damping', damp_new,
                'bandwidth', bw_new,
                'amplitude', self.amplitude,
                'time_vector', tVec,
                'parent', self.parent,
                'isForm', self.isForm)
        return obj_out

    def interpolate(self, txx):
        """ Interpolation of formant trajectory along the time axis """
        for attr in ["frequency", "damping", "bandwidth", "amplitude"]:
            f1 = scintp.interp1d(self.time_vector, getattr(self, attr), 
                                 kind='linear', axis=1, 
                                 bounds_error=False, fill_value=0)
            setattr(self, attr, f1(txx))       
        self.time_vector = txx
        self.extrap()
        
    def extrap(self):
        
        nb_formant = self.frequency.shape[0]
        for k in range(nb_formant):
            self.frequency[k,:] = fill_nan_and_zero(self.frequency[k,:])
            self.damping[k,:] = fill_nan_and_zero(self.damping[k,:])
            self.bandwidth[k,:] = fill_nan_and_zero(self.bandwidth[k,:])
            self.amplitude[k,:] = fill_nan_and_zero(self.amplitude[k,:])
        
    def create_spectral_envelope(self, nfft=None):
        
        if nfft is None:
            if hasattr(self.parent, 'sampling_frequency'):
                nfft = 2**nextpow2(self.parent.sampling_frequency)
            else:
                raise ValueError('Parent of formant object is None')
                
        frequency = self.frequency
        damping = self.damping
        sr = self.parent.sampling_frequency
        L = int(nfft / 2 + 1)
        nb_frame = len(self.time_vector)
        
        z_pos = np.exp(2 * np.pi * 1j * frequency / sr) * np.exp(-damping / sr)
        z_whole = np.concatenate((z_pos, np.conj(z_pos)), 0)
        
        spec_envelope = np.zeros((L, nb_frame), dtype='complex')
        
        for k in range(nb_frame):
            fxx, h_curr = z2freqz(1, z_whole[:,k], sr=sr, nfft=nfft)    
            spec_envelope[:,k] = h_curr
            
        self.parent.spectral_envelope = TimeFreq('time_frequency_matrix', spec_envelope,
                                                 'time_vector', self.time_vector,
                                                 'frequency_vector', fxx, 
                                                 'parent', self.parent)

    def get_stats(self, attr='frequency'):
        self.stats = []
        x = getattr(self, attr)
        for k in range(4):
            obj = RandomVar('samples', x[k,:])
            obj.get_stats(50)
            self.stats.append(obj)

    def plotformantspace(self, n1=1, n2=2, h=None):
        print('Plotting formant space')
        if n1 is not None and n2 is not None:
            h, ax1 = plottools.vowel_space(self.frequency, n1=1, n2=2)
        else:
            if self.stats is None:
                self.get_stats()
            return plottools.formantspace(self, subax=[5, 9, 10, 13, 14, 15],
                             h=None, clr='b')

    def compareformantspace(self, objForm):

        h = self.plotformantspace(n1=None, n2=None)
        if objForm.stats is None:
            objForm.get_stats()
        h = plottools.formantspace(objForm, subax=[2, 3, 4, 7, 8, 12],
                              h=h, clr='r', order=2)

    def plotstats(self):
        return plottools.formant_stats(self)

    def comparestats(self, objForm):
        return plottools.formant_stats(self, objForm)

    def comparehull(self, objForm):
        return plottools.hull_compare(self, objForm)
    
class MFCC:
    
    coeff = None
    nb_coeff = None
    time_vector = None
    mel_vector = None
    frequency_vector = None
    power_log = None
    parent = None
    c0 = None
    energy = None
    
    def __init__(self, *args):
        nargin = len(args)
        for k in range(0,nargin,2):
            key, value = args[k:k+2]
            setattr(self, key, value)
        
    def copy(self):
        return copy.deepcopy(self)

    def mfcc2melspectro(self, freq_out=None):
        
        nb_coeff = self.nb_coeff + 1
        coeff = self.coeff[:nb_coeff,:]
        nb_filt = len(self.mel_vector)
        if self.c0 is not None:
            coeff[0,:] = [x for x in self.c0] 
        # basis  = dct(nb_coeff + 1, filter_len=nb_filt-2)[1:,:]
        basis  = dct(nb_coeff, filter_len=nb_filt-2)
        pxx_new = np.linalg.pinv(basis) @ coeff
        
        S_log = TimeFreq('time_vector', self.time_vector, 
                         'frequency_vector', self.mel_vector[1:-1], 
                          'time_frequency_matrix', np.abs(10**(pxx_new/10)),
                          'parent', self.parent)

        self.power_log = pxx_new
        mels_lin = np.linspace(0, self.mel_vector[-1], len(self.mel_vector))
        S_log.interpolate(mels_lin, axis='freq')
        freq_mel = mel2freq(mels_lin)
        S_log.frequency_vector = freq_mel
        freq_lin = np.linspace(0, freq_mel[-1], len(freq_mel) )
        S_log.interpolate(freq_lin, axis='freq')
        if freq_out is not None:
            S_log.interpolate(freq_out, axis='freq')
        S_log.time_frequency_matrix = np.abs(S_log.time_frequency_matrix)
        if self.parent is not None:
            self.parent.spectral_envelope = S_log
    
        return S_log
    
class Speaker:

    mother_tongue = None
    speaker_id = None
    L2 = None
    L2_level = None
    gender = None
    age = None
    utterance = []
    phonemes = []
    formants_stats = None
    pitch_stats = None
    intensity_stats = None
    formants = None
    accent = None

    def __init__(self, *args):
        nargin = len(args)
        for k in range(0,nargin,2):
            key, value = args[k:k+2]
            setattr(self, key, value)

    def get_all_formants(self, attr='utterance', meth=None):
        """ returns formants values of all utterances in one Formant object """

        y = []
        y_bw = []
        list_attr = getattr(self, attr)
        for numk, k in enumerate(list_attr):
            nwin = 512
            novlp = int(nwin * 3 / 4)
            Fr = k.formants
            n_frame = Fr.frequency.shape[1]
            if n_frame > 0:
                if Fr is None:
                    Fr = k.formantestimate(nwin, novlp, ord_filt = 5)
                if meth == 'mean':
                    Fr.frequency = np.nanmean(Fr.frequency, axis=1).reshape(-1,1)
                    Fr.bandwidth = np.nanmean(Fr.bandwidth, axis=1).reshape(-1,1)
                if meth == 'median':
                    Fr.frequency = np.nanmedian(Fr.frequency, axis=1).reshape(-1,1)
                    Fr.bandwidth = np.nanmedian(Fr.bandwidth, axis=1).reshape(-1,1)
                if meth == 'central':
                    Fr.frequency = Fr.frequency[:,int(n_frame/2)].reshape(-1,1)
                    Fr.bandwidth = Fr.bandwidth[:,int(n_frame/2)].reshape(-1,1)
                if numk == 0:
                    y = Fr.frequency
                    y_bw = Fr.bandwidth
                else:
                    y = np.hstack((y, Fr.frequency))
                    y_bw = np.hstack((y_bw, Fr.bandwidth))
            else:
                pass
            
        y, idx = utils.reject_outliers(y)
        y_bw = y_bw[:,idx]            
        self.formants = Formants('frequency', y,
                                 'bandwidth', y_bw,
                                 'parent', self)

    def get_stats(self, attr):

        setattr(self, attr, [])

        for numk, k in enumerate(self.utterance):

            if attr != 'formant_stats':
                if attr == 'pitch_stats':
                    pros_feat = k.getpitch()
                if attr == 'intensity_stats':
                    pros_feat = k.getintensity()
                prmTmp = ProsodyParam('mean', np.nanmean(pros_feat.values))
                prmTmp.median = np.nanmedian(pros_feat.values)
                pros_feat.getcontour()
                cont = pros_feat.contour
                cont.values = pros_feat.values
                cont.getbaselines()
                prmTmp.declination = cont.declination
                prmTmp.span = cont.getpitchspan()
                prmTmp.peak_extent = cont.getnppe()
                prmTmp.local_peak_dynamic = cont.getlpdyn()
            else:
                formants = k.formantestimate(512, 384, ord_filt=5)
                freq = formants.frequency
                t_form = formants.time_vector
                prmTmp = ProsodyParam('mean', np.nanmean(freq, 1),
                                                   'raw_values', freq,
                                                   'span', [],
                                                   'declination', [],
                                                   'peak_extent', [],
                                                   'local_peak_extent', [])
                prmTmp.median = np.nanmedian(freq, 1)
                for k in range(4):
                    cont = Contour('values', freq[k,:].squeeze(),
                                  'time_points', t_form.squeeze())
                    cont.getbaselines()
                    prmTmp.declination.append(cont.declination)
                    prmTmp.span.append(cont.getpitchspan())
                    prmTmp.peak_extent.append(cont.getnppe())
                    prmTmp.local_peak_dynamic.append(cont.getlpdyn())

            getattr(self, attr).append(prmTmp)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jun 25 07:54:59 2020

@author: benjamin
"""

import numpy as np
import numpy.matlib
import scipy.signal
import warnings
with warnings.catch_warnings():
    warnings.filterwarnings("ignore",category=FutureWarning)
    import statsmodels.api as sm
import h5py
from sklearn import linear_model
from .prosody import Contour
from .ola import *
from joblib import Parallel, delayed

def findelements(a, b):
    # find indices of elements in a that contains elements in b
    nB = len(b)
    idx = []
    
    for k in range(nB):
        bTmp = b[k]
        idxTmp = [i for i,x in enumerate(a) if x==bTmp]
        if idxTmp != []: 
            if idx == []:
                idx = idxTmp
            else:
                idx = np.concatenate((idx, idxTmp)).astype(int)
                
    return idx

def get_distribution_from_keys(f, key):
    
    y = f[key][()]
    pdf = []
    pdf_bins = []
    
    if len(y) == 3:
        y = y[0]
    
    if len(y) == 4:    
        for k in y:
            pdfTmp, bins_tmp = np.histogram(k, 25, density=True)
            pdf.append(sm.nonparametric.lowess(pdfTmp, np.arange(25),
                                          frac=0.2,
                                          return_sorted=False))
            pdf_bins.append(bins_tmp)
    else:
        pdfTmp, bins_tmp = np.histogram(y, 25, density=True)
        pdf.append(sm.nonparametric.lowess(pdfTmp, np.arange(25),
                                      frac=0.2,
                                      return_sorted=False))
        pdf_bins.append(bins_tmp)
    return pdf, pdf_bins

def get_distribution_from_file(dist_file):    
    
    f = h5py.File(dist_file, 'r')
    a = dict.fromkeys(f.keys(), 0)
    for k in f.keys():        
        a[k] = get_distribution_from_keys(f, k)
        
    return a

def get_pros_modif_factor_from_dist(dist):
    fact_med = random_smpl(dist, 'med_pitch', lim=[0.5,2])
    fact_span = random_smpl(dist, 'span_pitch', lim=[0.25,3])
    fact_dur = random_smpl(dist, 'duration', lim=[1/1.8,1.8])
    return fact_med, fact_span, fact_dur

def parallel_lasso_fun(frame, A_source, features, alpha=1e-3):
    feature_frame = features[:, frame]
    clf = linear_model.Lasso(alpha=alpha, tol=1e-10, max_iter=100000,
                                 positive=True)    
    clf.fit(A_source, feature_frame)
    return clf.coef_

def parallel_lasso(A_source, features, n_jobs, alpha=1e-3):
    inputs = range(features.shape[1])
    pList = Parallel(n_jobs=n_jobs, prefer="threads")(delayed(parallel_lasso_fun)(k, 
                                          A_source, features, alpha=alpha) for k in inputs)
    return np.array(pList).T
    

def parse_align_file(align_file):
    f = open(align_file, 'r', encoding='Latin 1')
    lines = f.readlines()
    keys = ['utterance_id', 'channel', 'start', 'dur', 'phone']
    rlist = {"utterance_id":[], "channel":[], "start":[], "dur":[], "phone":[]}
    
    for curr_line in lines:
        for ks, curr_split in enumerate(curr_line.split(' ')):
            rlist[keys[ks]].append(curr_split.replace('\n', ''))
    return rlist

def parse_txt_file(txt_file):
    f = open(txt_file, 'r', encoding='Latin 1')
    lines = f.readlines()
    utt_id = [x[:x.find(' ')] for x in lines]
    utt_txt = [x[x.find(' ') + 1:] for x in lines]
    return utt_id, utt_txt

def parse_rlist_file(rlist_file):
    f = open(rlist_file, 'r', encoding='Latin 1')
    lines = f.readlines()
    keys = ['speaker_id', 'utterance_id', 'path_to_utterance', 
            'channel', 'start', 'end']
    rlist = {"speaker_id":[], "utterance_id":[], "path_to_utterance":[], 
             "channel":[], "start":[], "end":[]}
    
    for curr_line in lines:
        for ks, curr_split in enumerate(curr_line.split(' ')):
            rlist[keys[ks]].append(curr_split.replace('\n', ''))
    return rlist

def parse_slist_file(slist_file):
    f = open(slist_file, 'r', encoding='Latin 1')
    lines = f.readlines()
    keys = ['speaker_id', 'utterance_id', 'path_to_utterance', 
            'path_to_segmentation']
    slist = {"speaker_id":[], "utterance_id":[], "path_to_utterance":[], 
             "path_to_segmentation":[]}
    
    for curr_line in lines:
        for ks, curr_split in enumerate(curr_line.split(' ')):
            slist[keys[ks]].append(curr_split.replace('\n', ''))
    return slist

def random_its(pdf, pdf_bins, N=1):
    
    x = np.linspace(min(pdf_bins[1:]), max(pdf_bins[1:]), 10000)
    f1 = scipy.interpolate.interp1d(pdf_bins[1:], pdf)
    y = f1(x)
    
    cdf = np.cumsum(y/sum(y))
    R = np.random.uniform(0, 1, N)
    return [x[np.argwhere(cdf == min(cdf[(cdf - r) > 0]))[0][0]] for r in R]

def random_smpl(dist, key, k=0, lim=None):
    
    pdf = dist[key][0][k]
    pdf_bins = dist[key][1][k]
      
    fact = random_its(pdf, pdf_bins, 1)[0]
    if lim is not None:
        while fact > lim[1] or fact < lim[0]:
            fact = random_its(pdf, pdf_bins, 1)[0]
    return fact

def read_dictionary_file(dict_file):
    
    f = h5py.File(dict_file, 'r')
    A_dict = f['dictionary'][()]
    phones = f['phonemes'][()]
    info = f["type"][()].decode("utf-8")
    
    return A_dict, phones, info

def read_spk_model(model):
    
    f = h5py.File(model, 'r')
    W_mask = f['mask_dict'][()]
    W_nominal = f['nominal_dict'][()]
    
    return W_nominal, W_mask

def sparse_regression(features, A_source, A_target, is_norm=True, is_res=True, 
                      n_jobs=1, alpha=1e-3, feat_type="log"):    
    
    feat_type = "log"
    if is_norm:
        b_m = np.mean(features, axis=0, keepdims=True)
        b_s = np.std(features, axis=0, keepdims=True)
        b_s[b_s==0] = 1
        if feat_type == "lin":
            b_m = 0
        features = (features - b_m) / b_s
        if feat_type == "log":
            A_source = ((A_source - np.mean(A_source, axis=0, keepdims=True)) / 
                        np.std(A_source, axis=0, keepdims=True))
            A_target = ((A_target - np.mean(A_target, axis=0, keepdims=True)) /
                        np.std(A_target, axis=0, keepdims=True))
        else:
            A_source = A_source / np.std(A_source, axis=0, keepdims=True)
            A_target = A_target / np.std(A_target, axis=0, keepdims=True)
    else:
        b_m = 0
        b_s = 1    
    
    if n_jobs == 1:
        clf = linear_model.Lasso(alpha=alpha, tol=1e-10, max_iter=100000,
                                 positive=True)
        clf.fit(A_source, features)
        W = clf.coef_.T
    else:
        W = parallel_lasso(A_source, features, n_jobs, alpha=alpha)
    if is_res:
        Rs = features - A_source @ W
        Tk = regul_res(A_source, A_target)
        S_target = warp_res(A_target, W, Tk, Rs)
    else:
        S_target = A_target @ W
    # res = ((features - A_source @ W) * is_res)
    return (S_target)*b_s + b_m

def transform_sparse_audio(Sp, W_source, W_target, output_audio, isFlat=False,
                           isNorm=False, is_feat_norm=True, is_residual=False, 
                           feat_type="lin", n_jobs=1, alpha=1e-3):
    
    Sp.sparse_transform(W_source, W_target, isFlat=isFlat, isNorm=isNorm, 
                        is_feat_norm=is_feat_norm, is_residual=is_residual, 
                        feat_type=feat_type, n_jobs=n_jobs, alpha=alpha)
    Sp.savespeech(output_audio, "normalized")
    
    return len(Sp.signal) / Sp.sampling_frequency
   
def transform_formant(formants, dil_fact, trans_fact=np.ones(4)):  
    fr_out = formants.copy()
    f_mean = np.mean(formants.frequency, 1).squeeze()   
    detrend_freq = ((fr_out.frequency - f_mean.reshape(-1,1)) *
                        (1 - dil_fact.reshape(-1,1)))
    trans_freq = (f_mean * trans_fact).reshape(-1,1)
    new_freq = detrend_freq + trans_freq
    fr_out.frequency = new_freq  
    return fr_out

def transform_formant_from_dist(Fr, dist, rndmize=True):
    
    Fr_out = Fr.copy()
    for k in range(4):
        
        if rndmize:
            fact_med = random_smpl(dist, 'med_formant', k=k, lim=[0.5,2])
            fact_span = random_smpl(dist, 'span_formant', k=k, lim=[0.25,3])
        else:
            fact_med = dist["med_formant"][k]
            fact_span = dist["span_formant"][k]
        
        fTmp = Fr_out.frequency[k,:]
        Cont = Contour('values', fTmp, 'time_points', Fr_out.time_vector)
        try:
            Cont.getbaselines()
            Cont.changespan(fact_span)
        except:
            Cont = Contour('values', fTmp, 'time_points', Fr_out.time_vector)
                
        Cont.values *= fact_med 
        Fr_out.frequency[k,:] = Cont.values
    
    return Fr_out

def transform_pitch(pitch, fact_med, fact_span):
    
    pitch_out = pitch.copy()    
    pitch_out.getcontour()
    if fact_span != 1:
        try:
            pitch_out.changespan(fact_span)
        except:
            pitch_out = pitch.copy()
            pitch_out.getcontour()
    if fact_med != 1:
        pitch_out.values *= fact_med
     
    return pitch_out

def regul_res(A_source, A_target):
    nb_feat, nb_ph = A_source.shape
    Tk = []
    for k in range(nb_ph):
        As_tmp = A_source[:,k].reshape(-1,1)
        At_tmp = A_target[:,k].reshape(-1,1)
        Tk.append(At_tmp @ np.linalg.pinv(As_tmp))

    return Tk

def warp_res(A_targ, W_s, T_k, R_s):
    nb_feat, nb_ph = A_targ.shape
    nb_frame = R_s.shape[1]
    X_t = np.zeros_like(R_s)
    T_k.append(np.eye(nb_feat))
    W_plus = np.vstack((W_s, np.ones((1, nb_frame))))

    for k_fr in range(nb_frame):
        s = 0
        for k_ph in range(nb_ph + 1):
            s += W_plus[k_ph, k_fr] * T_k[k_ph]
        X_t[:, k_fr] = A_targ @ W_s[:, k_fr] + s @ R_s[:, k_fr]

    return X_t


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct  6 09:32:12 2020

@author: benjamin
"""

class AcousticsElem:
    """ Values of electric components of the equivalent circuit""" 
    
    parent = None
    Lj = None
    Rj = None
    Cj = None
    Gw = None
    wl = None
    wc = None
    Udj = None
    bj = None
    Hj = None
    Rcm = None
    Rcp = None
    Ns = None
    
    def __init__(self, *args):
        nargin = len(args)
        for k in range(0, nargin, 2):
            key, value = args[k:k+2]
            setattr(self, key, value)
            
class DerivTerms:
    """ contains the derivaztive terms"""
    
    parent = None
    Q = None
    Qwl = None
    Qwc = None
    Qnc = None

    def __init__(self, *args):
        nargin = len(args)
        for k in range(0, nargin, 2):
            key, value = args[k:k+2]
            setattr(self, key, value)
            
class IntegrTerms:
    """ Contains the integration terms"""
    
    parent = None
    Vc = None
    V = None
    
    def __init__(self, *args):
        nargin = len(args)
        for k in range(0, nargin, 2):
            key, value = args[k:k+2]
            setattr(self, key, value)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Feb  6 21:31:52 2019

@author: benjamin
"""

from .utils import *
# from .signaltools import *
from .vtnetwork import VtNetwork 
from .vtwaveguide import *
from .subclasses import *
from .oscillator import *
from .speechaudio import *
from ._set import *
from .jsontools import *
from .filetools import *
from .prosody import *
from .areafunction import *
from .electricterms import *
from .acoustics import *
from .plottools import *
from .stats import *
from .ola import *
from .audiotools import *
from .inittools import *
from .praattools import *
from .transformtools import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jun 25 07:54:59 2020

@author: benjamin
"""

import soundfile
import numpy as np
import copy


def record_to_file(path, data, sample_width, rate):    
    """ saves audio sigbal into file """
    soundfile.write(path, data, int(rate))

def clipper(x, threshold, method, is_relative=True):
    """ returns a soft or hard clipped version of an audio signal"""
    abs_thresh = threshold
    if is_relative:
        abs_thresh = np.max(abs(x)) * threshold

    y = copy.deepcopy(x)
    if method.lower() == 'hard':
        y[x >= abs_thresh] = abs_thresh
        y[x <= -abs_thresh] = -abs_thresh
    if method.lower() == 'soft':
        abs_thresh2 = abs_thresh
        y[x >= abs_thresh2] = abs_thresh
        y[x <= -abs_thresh2] = -abs_thresh
        idx = [xTmp for xTmp in range(len(x)) if
               x[xTmp] >= abs_thresh2 and x[xTmp] <= -abs_thresh2]
        y[idx] = y[idx] - (y[idx]**3) / 3
    if method.lower() not in ['hard', 'soft']:
        raise ValueError("Unexpected method for distortion. Use either hard or soft")
    return y



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct  6 09:32:12 2020

@author: benjamin
"""

import copy
import numpy as np
from ._set import timestruct, struct
from .electricterms import AcousticsElem, DerivTerms, IntegrTerms

class AcousticsTime:
    """ contains acoustic terms of the vocal tract """

    parent = None
    pressure_freq = None
    flow_freq = None
    areasensitivity = None
    lengthsensitivity = None
    norder = 1
    pressure_time = None
    flow_time = None
    elements = None
    derivative_terms = None
    intergation_terms = None
    noise_matrix = None
    flow_time_matrix = None
    pressure_time_matrix = None

    def __init__(self, *args):
        nargin = len(args)
        for k in range(0, nargin, 2):
            key, value = args[k:k+2]
            setattr(self, key, value)

    def copy(self):
        """ returns a deep copy of the instance"""

        return copy.deepcopy(self)

    def initderiv(self):
        """ Initialisation of the derivative and integration terms
        Nx is the number of tubelets of the considered pipe"""

        nTube = self.parent.area_function.area.shape[0]
        Vc = np.zeros(nTube)
        Q = np.zeros(nTube + 1)
        V = np.zeros(nTube + 1)
        Qwl = np.zeros(nTube)
        Qwc = np.zeros(nTube)
        if self.parent.child_wvg is not None:
            nChild = len(self.parent.child_wvg)
        else:
            nChild = 0
        Qnc = np.zeros(nChild)

        deriv_terms = DerivTerms('Q', Q,
                                 'Qwl', Qwl,
                                 'Qwc', Qwc,
                                 'Qnc', Qnc,
                                 'parent', self.parent)
        integr_terms = IntegrTerms('V', V,
                                   'Vc', Vc,
                                   'parent', self.parent)

        self.derivative_terms = deriv_terms
        self.integration_terms = integr_terms

    def acou2elec(self, Const=None, ujm1=0, aglot=0, isSub=False):
        """Computes the electric components values of the vocal tract"""
        
        if Const is None:
            Const = timestruct()

        Xj = self.parent.area_function.length
        Aj = self.parent.area_function.area

        if hasattr(Const, 'k_iter'):
            Xj = Xj[:, Const.k_iter].reshape(-1, 1)
            Aj = Aj[:, Const.k_iter].reshape(-1, 1)

        nTube, nFrame = Aj.shape
        A_lip = Aj[-1, :]
        rad = np.sqrt(Aj/np.pi)
        aprec = np.diff(np.append(aglot, Aj.squeeze()))
        asucc = np.sign(np.diff(np.append(Aj.squeeze(), 1e10)))

        Gw = np.zeros(Xj.shape)
        Udj = np.zeros(Xj.shape)
        Rcm = np.zeros(Xj.shape)
        Rcp = np.zeros(Xj.shape)

        if Const.dynterm and sum(abs(ujm1)) != 0:
            idx = [x for x in range(len(aprec)) if aprec[x] < 0]
            Rcm[idx, :] = abs(ujm1[idx].reshape(-1, 1))*Const.rho/(2*Aj[idx, :]**2)
            idx = [x for x in range(len(asucc)) if asucc[x] < 0]
            idx2 = [x + 1 for x in idx]
            Rcp[idx, :] = -abs(ujm1[idx2].reshape(-1, 1))*Const.rho/(2*Aj[idx, :]**2)

        Lj = 0.5*Const.rho*Xj/Aj
        Rj = Const.rmlt * np.pi * Const.mu * Xj / Aj**2
        Cj = Xj*Aj/(Const.rho*Const.c**2)
        Srad = 0.5*Const.T*Const.S*np.pi*np.sqrt(np.pi)/Const.rho
        Grad = Const.G*np.pi**2/Const.rho/Const.c
        if isSub:
            xtrach = np.cumsum(Aj) / sum(Aj)
            wr = 0.33/(1.30**(xtrach/4))*1.6e3/(2*np.pi*(rad**3)*Xj)
            wl = 0.33/(1.30**(xtrach/4))*1.1e-3/(2*np.pi*rad*Xj)
            wc = (2*np.pi*(rad**3)*Xj)/(0.33/(1.30**(xtrach/4))*0.392e6)
        else:
            wr = Const.wr/(2*np.sqrt(np.pi))/Xj/np.sqrt(Aj)
            wl = 2/Const.T*Const.wm/(2*np.sqrt(np.pi))/Xj/np.sqrt(Aj)
            wc = 0.5*Const.T/(Const.wc/(2*np.sqrt(np.pi))/Xj/np.sqrt(Aj))

        if Const.wallyield:
            Gw = 1 / (wr + wl + wc)

        #### Compute linear system coefficients
        bj = np.zeros((nTube+1, nFrame))
        Hj = np.zeros((nTube+1, nFrame))
        bj[0:nTube, :] = 1./(2*Cj/Const.T+Gw)
        Hj[1:nTube, :] = -2*(Lj[0:nTube-1, :]+Lj[1:nTube, :])/Const.T \
            -Rj[0:nTube-1, :]-Rj[1:nTube, :]-Rcm[0:nTube-1, :]-Rcp[1:nTube, :] \
            -bj[0:nTube-1, :]-bj[1:nTube, :]
        bj[nTube, :] = 1./(Srad*np.sqrt(A_lip)+Grad*A_lip)
        Hj[nTube, :] = -(2/Const.T*Lj[nTube-1, :]+Rj[nTube-1, :] + \
                         bj[nTube-1, :]+bj[nTube, :]+Rcp[nTube-1])
        elements = AcousticsElem('Lj', Lj.squeeze(),
                                 'Rj', Rj.squeeze(),
                                 'Cj', Cj.squeeze(),
                                 'Gw', Gw.squeeze(),
                                 'wl', wl.squeeze(),
                                 'wc', wc.squeeze(),
                                 'Udj', Udj.squeeze(),
                                 'bj', bj.squeeze(),
                                 'Hj', Hj.squeeze(),
                                 'Rcm', Rcm.squeeze(),
                                 'Rcp', Rcp.squeeze(),
                                 'Ns', np.zeros((nTube+1, nFrame)).squeeze(),
                                 'parent', self)
        self.elements = elements

class AcousticsFreq:
    """ contains the acoustics terms of the vocal tract in the frequency domain"""

    parent = None
    pressure_freq = None
    flow_freq = None
    areasensitivity = None
    lengthsensitivity = None

    def __init__(self, *args):
        nargin = len(args)
        for k in range(0, nargin, 2):
            key, value = args[k:k+2]
            setattr(self, key, value)

    def copy(self):
        """ returns a deep copy of the instance"""

        return copy.deepcopy(self)

    def af2acoustics(self, freq, param=None):
        """Compute the pressure and air flow inside the vocal and nasal tracts
        defined by area function A and l.
        The constant values are stocked in param.
        After the chain matrix paradigm by Sondhi and Schroeter
        in "A hybrid time-frequency domain articulatory speech synthesizer",
        IEEE TASSP, 1987)."""

        if param is None:
            param = struct()

        af = self.parent.area_function.area
        lf = self.parent.area_function.length
        if len(af.shape) == 1:
            nFrame = 1
            af = af.reshape((af.shape[0], nFrame), order='F')
            lf = lf.reshape((af.shape[0], nFrame), order='F')
        else:
            nFrame = af.shape[1]
        shapeFreq = freq.shape
        nFreq = shapeFreq[0]
        if len(shapeFreq) == 1:
            freq = np.matlib.repmat(freq, nFreq, nFrame)

        w = 2 * np.pi * freq
        alp = np.sqrt(1j*w * param.c1)
        bet = 1j*w * param.wo2/((1j*w + param.a) * 1j*w + param.b) + alp
        gam = np.sqrt((param.alp + 1j*w) / (bet + 1j*w))
        sig = gam * (bet + 1j*w)

        P = np.zeros((af.shape[0], nFrame, nFreq)) + 1j*np.zeros((af.shape[0], nFrame, nFreq))
        U = np.zeros((af.shape[0], nFrame, nFreq)) + 1j*np.zeros((af.shape[0], nFrame, nFreq))

        for kframe in range(nFrame):

            af_tmp = af[:, kframe]
            lf_tmp = lf[:, kframe]
            freqtmp = freq[:, kframe]
            sigtmp = sig[:, kframe]
            gamtmp = gam[:, kframe]
            wTmp = 2*np.pi*freqtmp

            Grad = 0.5 * param.rho * wTmp**2 / np.pi / param.c_s
            Jrad = 8 * param.rho * wTmp / 3 / np.pi**(1.5) * af_tmp[-1]**(-0.5)
            Zrad = Grad + 1j*Jrad
            P1 = Zrad
            U1 = np.ones(P1.shape)
            for ktube in range(af.shape[0]-1, -1, -1): # Compute the chain matrix
                Argh = sigtmp * lf_tmp[ktube] / param.c_s
                A = np.cosh(Argh)
                B = -param.rho*param.c_s / af_tmp[ktube] * gamtmp * np.sinh(Argh)
                C = -af_tmp[ktube] / param.rho/param.c_s / gamtmp * np.sinh(Argh)
                det_k = 1 / (A**2 - B * C)
                P1tmp = (A * P1 - B * U1) / det_k
                U1tmp = (-C * P1 + A * U1) / det_k
                P[ktube, kframe, :] = P1tmp.reshape((1, 1, np.size(P1tmp)),
                                                    order='F')
                U[ktube, kframe, :] = U1tmp.reshape((1, 1, np.size(U1tmp)),
                                                    order='F')
                P1 = P1tmp
                U1 = U1tmp

        self.pressure_freq = P
        self.flow_freq = U
        return P, U

    def areasensitivityfunction(self, freq, param=struct()):
        """Compute the sensitivity function Sn along the vocal tract"""

        af = self.parent.area_function.area
        lf = self.parent.area_function.length
        if len(af.shape) == 1:
            nFrame = 1
            af = af.reshape((af.shape[0], nFrame), order='F')
            lf = lf.reshape((af.shape[0], nFrame), order='F')
        else:
            nFrame = af.shape[1]
        nFreq, _ = freq.shape

        Sn_x = np.zeros((af.shape[0], nFrame, nFreq * nFrame)) + \
            1j*np.zeros((af.shape[0], nFrame, nFreq * nFrame))

        P, U = self.af2acoustics(freq, param)

        for kframe in range(nFrame):
            for kfreq in range(nFreq):
                Pn = np.squeeze(abs(P[:, kframe, kfreq])**2)
                Un = np.squeeze(abs(U[:, kframe, kfreq])**2)
                lftmp = np.squeeze(lf[:, kframe])
                aftmp = np.squeeze(af[:, kframe])
                idx = kframe*nFreq+kfreq
                cst1 = param.rho/aftmp*Un-aftmp/param.rho/param.c_s**2*Pn
                argsum1 = param.rho/aftmp*Un+aftmp/param.rho/param.c_s**2*Pn
                Sn_x[:, kframe, idx] = lftmp * cst1 / np.sum(lftmp * argsum1)

        self.areasensitivity = Sn_x
        return Sn_x

    def lengthsensitivityfunction(self, freq, param):
        """Compute the sensitivity function Sn along the vocal tract for the
        length function"""

        af = self.parent.area_function.area
        lf = self.parent.area_function.length

        if len(af.shape) == 1:
            nFrame = 1
            af = af.reshape((af.shape[0], nFrame),
                            order='F')
            lf = lf.reshape((af.shape[0], nFrame),
                            order='F')
        else:
            nFrame = af.shape[1]

        Sn_x = np.zeros((af.shape[0], nFrame, freq.shape[0] * nFrame)) + \
            1j*np.zeros((af.shape[0], nFrame, freq.shape[0] * nFrame))
        rho = param.rho
        cs = param.c_s

        P, U = self.af2acoustics(freq, param)
        for kframe in range(nFrame):
            for kfreq in range(freq.shape[0]):
                Pn = np.squeeze(abs(P[:, kframe, kfreq])**2)
                Un = np.squeeze(abs(U[:, kframe, kfreq])**2)
                lftmp = np.squeeze(lf[:, kframe])
                aftmp = np.squeeze(af[:, kframe])
                idx = kframe * freq.shape[0] + kfreq
                Sn_x[:, kframe, idx] = lftmp*(rho/aftmp*Un+aftmp/rho/cs**2*Pn)/ \
                    np.sum(lftmp*(rho/aftmp*Un+aftmp/rho/cs**2*Pn))

        self.lengthsensitivity = Sn_x
        return Sn_x



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jun 25 07:54:59 2020

@author: benjamin
"""

from os import path
from .speechaudio import SpeechAudio

import numpy as np

def file2speech(fileName, chn=1, sro=None, start=0, end=-1):
    
    if path.isfile(fileName):
        _, ext = path.splitext(fileName)
        if ext in [".mp3", ".sph", ".flac"]:
            import librosa
            import warnings
            warnings.filterwarnings('ignore')
            y, sr = librosa.load(fileName)
        else:
            import soundfile 
            y, sr = soundfile.read(fileName)
        if len(y.shape) == 2:
            idx = np.argmax(y.shape)
            if idx == 0:
                y = y[:,chn-1]
            else:
                y = y[chn-1,:]
                
        idx1 = 0
        idx2 = len(y)
        if start > 0:
            idx1 = int(start * sr)
        if end > 0:
            idx2 = int(end * sr)            
        if idx1 > 0 or idx2 < len(y):
            y = y[idx1:idx2]
        try:
            SpeechObject = SpeechAudio('signal', y.astype(float), 
                                       'sampling_frequency', sr)
        except:
            SpeechObject = SpeechAudio('signal', np.float32(y), 
                                       'sampling_frequency', sr)
        del y
            
        if sro is not None:
            if sr != sro:
                SpeechObject.speechresample(sro)
    else:
        raise ValueError("File " + fileName + " does not exist")    
    return SpeechObject


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Feb  7 10:47:40 2019

@author: benjamin
"""

from setuptools import setup, find_packages

setup(name='spychhiker',
      version='0.8.13',
      description='Various python class for speech analysis and speech synthesis',
      url=' https://gitlab.com/benjamin.elie/spychhiker',
      author='Benjamin Elie',
      author_email='bnjmn.elie@gmail.com',
      license='CeCILL',
      packages=find_packages(),
      install_requires=[
          'h5py',
        'numpy',
        'matplotlib',
        'scipy',
        'statsmodels',
        'librosa',
        'praat-parselmouth',
        'scikit-learn',
        'tqdm'
    ],
      zip_safe=False)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun  2 14:52:12 2021

@author: benjamin
"""

import numpy as np

def check_list(x):
    if isinstance(x, float) or isinstance(x, int):
            x = [x]
    return x

def get_index(spline, time_points):
    return np.array([(np.abs(spline.knots - x)).argmin() - 2 for x in time_points])

def move_spline(spline, time_points, factor):
    
    idx = get_index(spline, time_points)
    spline.knots[idx] += factor

    idxSort = np.argsort(spline.knots)
    spline.knots = spline.knots[idxSort]
    spline.coeff = spline.coeff[idxSort]
    spline.interpolate()

def target_spline(spline, time_points, targets):
    
    idx = get_index(spline, time_points)
    idx_time = np.array([(np.abs(spline.time_points - x)).argmin() 
                         for x in time_points])
    curr_val = spline.values[idx_time]
    nb_iter = 0
    nb_pts = len(time_points)
    cost_fun = np.sum(np.abs(np.array(targets) - np.array(curr_val)))
    while cost_fun > nb_pts and nb_iter < 100:
        curr_factor = np.array(targets) / np.array(curr_val)        
        spline.coeff[idx] *= curr_factor
        spline.interpolate()
        curr_val = spline.values[idx_time]
        cost_fun = np.sum(np.abs(np.array(targets) - np.array(curr_val)))
        nb_iter += 1
        
    if nb_iter >= 100:
        print("Warning: number of iterations exceeds 100")
        
def translate_spline(spline, time_points, factor):
    
    idx = get_index(spline, time_points)
    spline.coeff[idx] *= factor
    spline.interpolate()